from typing import Any, Literal
from cmap import Color
from pydantic_compat import BaseModel, Field, field_validator
from pydantic import field_serializer, ValidationInfo

from himena.standards.model_meta import DimAxis
from himena.consts import PYDANTIC_CONFIG_STRICT
from himena.utils.misc import iter_subclasses


class StyledText(BaseModel):
    model_config = PYDANTIC_CONFIG_STRICT

    text: str = Field(..., description="Text content.")
    size: float | None = Field(None, description="Font size.")
    color: Any | None = Field(None, description="Font color.")
    family: str | None = Field(None, description="Font family.")
    bold: bool = Field(False, description="Bold style or not.")
    italic: bool = Field(False, description="Italic style or not.")
    underline: bool = Field(False, description="Underline style or not.")
    alignment: str | None = Field(None, description="Text alignment.")


class BasePlotModel(BaseModel):
    model_config = PYDANTIC_CONFIG_STRICT
    name: str = Field(default="", description="Name of the plot.")

    @classmethod
    def construct(cls, type: str, dict_: dict[str, Any]) -> "BasePlotModel":
        for subclass in iter_subclasses(BasePlotModel):
            if subclass.__name__.lower() == type:
                return subclass(**dict_)
        raise ValueError(f"Unknown plot type: {type!r}")

    @field_validator("name", mode="before")
    def _validate_name(cls, name: str) -> str:
        if name is None:
            return ""
        return name

    def model_dump_typed(self) -> dict[str, Any]:
        return {"type": type(self).__name__.lower(), **self.model_dump()}

    def plot_option_dict(self) -> dict[str, Any]:
        """Return the GUI option dict for this plot for editing."""
        return {"name": {"widget_type": "LineEdit", "value": self.name}}


class Axis(BaseModel):
    """Model that represents a plot axis."""

    model_config = PYDANTIC_CONFIG_STRICT

    lim: tuple[float, float] | None = Field(None, description="Axis limits.")
    scale: Literal["linear", "log"] = Field("linear", description="Axis scale.")
    label: str | StyledText | None = Field(None, description="Axis label.")
    ticks: Any | None = Field(None, description="Axis ticks.")
    grid: bool = Field(False, description="Show grid or not.")

    @field_validator("lim", mode="before")
    def _validate_lim(cls, lim) -> tuple[float, float] | None:
        if lim is None:
            return None
        _lim = tuple(lim)
        if len(_lim) != 2:
            raise ValueError(f"Must be a tuple of 2 floats but got: {lim!r}")
        return _lim


class AxesBase(BaseModel):
    """Layout model for an axes."""

    models: list[BasePlotModel] = Field(
        default_factory=list, description="Child plot models."
    )
    title: str | StyledText | None = Field(None, description="Title of the axes.")

    @field_serializer("models")
    def _serialize_models(self, models: list[BasePlotModel]) -> list[dict]:
        return [model.model_dump_typed() for model in models]

    # NOTE: The `mode` argument must be "plain" to avoid models falling back to
    # BasePlotModel instances.
    @field_validator("models", mode="plain")
    def _validate_models(cls, models: list):
        out = []
        for model in models:
            if isinstance(model, dict):
                model = model.copy()
                model_type = model.pop("type")
                model = BasePlotModel.construct(model_type, model)
            elif not isinstance(model, BasePlotModel):
                raise ValueError(f"Must be a dict or BasePlotModel but got: {model!r}")
            out.append(model)
        return out


class StackedAxesBase(BaseModel):
    """Layout model for stacked axes."""

    shape: tuple[int, ...]
    multi_dims: list[DimAxis] = Field(
        ..., description="Multi-dimensional axes of the plot stack."
    )
    models: dict[tuple[int, ...], list[BasePlotModel]] = Field(
        default_factory=list, description="Dict of child plot models."
    )
    title: str | StyledText | None = Field(None, description="Title of the axes.")

    @field_serializer("models")
    def _serialize_models(
        self,
        models: dict[tuple[int, ...], list[BasePlotModel]],
    ) -> dict[tuple[int, ...], list[dict]]:
        return {
            k: [model.model_dump_typed() for model in models]
            for k, models in models.items()
        }

    # NOTE: The `mode` argument must be "plain" to avoid models falling back to
    # BasePlotModel instances.
    @field_validator("models", mode="plain")
    def _validate_models(cls, models: dict[tuple[int, ...], list]):
        out = {}
        for k, models in models.items():
            _list = out[k] = []
            for model in models:
                if isinstance(model, dict):
                    model = model.copy()
                    model_type = model.pop("type")
                    model = BasePlotModel.construct(model_type, model)
                elif not isinstance(model, BasePlotModel):
                    raise ValueError(
                        f"Must be a dict or BasePlotModel but got: {model!r}"
                    )
                _list.append(model)
        return out

    @field_validator("multi_dims", mode="plain")
    def _validate_axes(cls, multi_dims, values: ValidationInfo) -> list[DimAxis]:
        shape = values.data["shape"]
        if multi_dims is None:
            return [DimAxis(name=f"axis-{i}") for i in range(len(shape))]
        if len(multi_dims) != len(shape):
            raise ValueError(
                f"Number of axes ({len(multi_dims)}) must match the number of dimensions "
                f"({len(shape)})"
            )
        return [DimAxis.parse(axis) for axis in multi_dims]


class HasColor(BaseModel):
    color: Any | None = Field(None, description="Color property.")

    @property
    def alpha(self) -> float | None:
        """Return the alpha value of the edge color."""
        if self.color is None:
            return None
        return Color(self.color).alpha

    @alpha.setter
    def alpha(self, value: float | None):
        """Set the alpha value of the edge color."""
        if self.color is None:
            return
        color = Color(list(Color(self.color).rgba[:3]) + [value])
        self.color = color
        return


class Face(HasColor):
    """Model for face properties."""

    hatch: Any | None = Field(None, description="Hatch pattern of the face.")


class Edge(HasColor):
    """Model for edge properties."""

    width: float | None = Field(None, description="Width of the edge.")
    style: Any | None = Field(None, description="Style of the edge.")


def parse_edge(kwargs: dict[str, Any]) -> dict:
    color = kwargs.pop("color", kwargs.pop("edge_color", None))
    width = kwargs.pop("width", kwargs.pop("edge_width", None))
    style = kwargs.pop("style", kwargs.pop("edge_style", None))
    alpha = kwargs.pop("alpha", kwargs.pop("edge_alpha", None))
    name = kwargs.pop("name", None)
    if kwargs:
        raise ValueError(f"Extra keyword arguments: {list(kwargs.keys())!r}")
    if alpha is not None:
        color = Color([*Color(color).rgba[:3], alpha])
    edge = Edge(color=color, width=width, style=style)
    return {"edge": edge, "name": name}


def parse_face_edge(kwargs: dict[str, Any]) -> dict:
    color = kwargs.pop("color", kwargs.pop("face_color", None))
    hatch = kwargs.pop("hatch", kwargs.pop("face_hatch", None))
    alpha = kwargs.pop("alpha", kwargs.pop("face_alpha", None))
    kwargs = parse_edge(kwargs)
    if kwargs.get("color") is None:
        kwargs["color"] = color
    if alpha is not None:
        color = Color([*Color(color).rgba[:3], alpha])
    face = Face(color=color, hatch=hatch)
    return {"face": face, **kwargs}



from typing import ClassVar, Literal, Sequence, TYPE_CHECKING, SupportsIndex

import numpy as np
from pydantic_compat import BaseModel, Field
from himena.standards.plotting import models as _m
from himena.standards.plotting.components import (
    Axis,
    AxesBase,
    StackedAxesBase,
    parse_edge,
    parse_face_edge,
)
from himena.consts import PYDANTIC_CONFIG_STRICT, StandardType

if TYPE_CHECKING:
    from typing import Self
    from numpy.typing import NDArray


class BaseLayoutModel(BaseModel):
    model_config = PYDANTIC_CONFIG_STRICT
    model_type: ClassVar[str] = StandardType.PLOT

    hpad: float | None = Field(None, description="Horizontal padding.")
    vpad: float | None = Field(None, description="Vertical padding.")
    hspace: float | None = Field(None, description="Horizontal space.")
    vspace: float | None = Field(None, description="Vertical space.")
    background_color: str = Field("#FFFFFF", description="Background color.")

    def merge_with(self, other: "BaseLayoutModel") -> "BaseLayoutModel":
        raise NotImplementedError

    def model_dump_typed(self) -> dict:
        return {"type": type(self).__name__.lower(), **self.model_dump()}

    @classmethod
    def construct(self, model_type: str, dict_: dict) -> "BaseLayoutModel":
        from himena.standards.plotting.layout3d import SingleAxes3D

        if model_type == "singleaxes":
            return SingleAxes.model_validate(dict_)
        if model_type == "row":
            return Row.model_validate(dict_)
        if model_type == "column":
            return Column.model_validate(dict_)
        if model_type == "grid":
            return Grid.model_validate(dict_)
        if model_type == "singleaxes3d":
            return SingleAxes3D.model_validate(dict_)
        raise ValueError(f"Unknown layout model type: {model_type!r}")

    def show(self) -> None:
        """Show the layout in the current himena window."""
        from himena.widgets import current_instance

        ui = current_instance()
        ui.add_object(self, type=self.__class__.model_type, title="Plot")
        return None


# NOTE: ModelsRef should not be a BaseModel because BaseModel copies lists on each
# construction, thus cannot be used as a reference for the internal plot models.
class ModelsRef:
    models: list[_m.BasePlotModel]

    def __init__(self, models):
        self.models = models

    def scatter(
        self,
        x: Sequence[float],
        y: Sequence[float] | None = None,
        *,
        symbol: str = "o",
        size: float | None = None,
        **kwargs,
    ) -> _m.Scatter:
        """Add a scatter plot model to the axes."""
        x, y = _norm_xy(x, y)
        model = _m.Scatter(
            x=x, y=y, symbol=symbol, size=size, **parse_face_edge(kwargs)
        )
        self.models.append(model)
        return model

    def plot(
        self, x: Sequence[float], y: Sequence[float] | None = None, **kwargs
    ) -> _m.Line:
        """Add a line plot model to the axes."""
        x, y = _norm_xy(x, y)
        model = _m.Line(x=x, y=y, **parse_edge(kwargs))
        self.models.append(model)
        return model

    def bar(
        self,
        x: Sequence[float],
        y: Sequence[float] | None = None,
        *,
        bottom: "float | Sequence[float] | NDArray[np.number] | None" = None,
        bar_width: float | None = None,
        orient: Literal["vertical", "horizontal"] = "vertical",
        **kwargs,
    ) -> _m.Bar:
        """Add a bar plot model to the axes."""
        x, y = _norm_xy(x, y)
        if bottom is None:
            bottom = 0
        model = _m.Bar(
            x=x, y=y, bottom=bottom, bar_width=bar_width, orient=orient,
            **parse_face_edge(kwargs),
        )  # fmt: skip
        self.models.append(model)
        return model

    def errorbar(
        self,
        x: Sequence[float],
        y: Sequence[float] | None = None,
        *,
        x_error: "float | Sequence[float] | NDArray[np.number] | None" = None,
        y_error: "float | Sequence[float] | NDArray[np.number] | None" = None,
        capsize: float | None = None,
        **kwargs,
    ) -> _m.ErrorBar:
        """Add an error bar plot model to the axes."""
        x, y = _norm_xy(x, y)
        model = _m.ErrorBar(
            x=x,
            y=y,
            x_error=x_error,
            y_error=y_error,
            capsize=capsize,
            **parse_edge(kwargs),
        )
        self.models.append(model)
        return model

    def band(
        self,
        x: Sequence[float],
        y0: Sequence[float],
        y1: Sequence[float] | None = None,
        *,
        orient: Literal["vertical", "horizontal"] = "vertical",
        **kwargs,
    ) -> _m.Band:
        """Add a band plot model to the axes."""
        x = np.asarray(x)
        y0 = np.asarray(y0)
        if y1 is None:
            y1 = np.zeros_like(y0)
        model = _m.Band(x=x, y0=y0, y1=y1, orient=orient, **parse_face_edge(kwargs))
        self.models.append(model)
        return model

    def span(
        self,
        start: float,
        end: float,
        *,
        orient: Literal["vertical", "horizontal"] = "horizontal",
        **kwargs,
    ) -> _m.Span:
        """Add a span plot model to the axes."""
        model = _m.Span(start=start, end=end, orient=orient, **parse_face_edge(kwargs))
        self.models.append(model)
        return model

    def hist(
        self,
        data: "Sequence[float] | NDArray[np.number]",
        *,
        bins: int = 10,
        range: tuple[float, float] | None = None,
        orient: Literal["vertical", "horizontal"] = "vertical",
        stat: Literal["count", "density", "probability"] = "count",
        **kwargs,
    ) -> _m.Histogram:
        """Add a histogram plot model to the axes."""
        match stat:
            case "count":
                data = data
                density = False
            case "density":
                data = data
                density = True
            case "probability":
                data = data / len(data)
                density = False
            case _:
                raise ValueError(f"Unsupported histogram stat: {stat}")
        height, bins = np.histogram(data, bins=bins, range=range, density=density)
        model = _m.Histogram(
            height=height, bins=bins, orient=orient, **parse_face_edge(kwargs),
        )  # fmt: skip
        self.models.append(model)
        return model

    def text(
        self,
        x: Sequence[float],
        y: Sequence[float],
        text: Sequence[str],
        *,
        size: int = 12,
        color: str = "black",
        family: str = "Arial",
        anchor: _m.ANCHOR_STRINGS = "center",
        rotation: float = 0,
    ) -> _m.Texts:
        """Add a text plot model to the axes."""
        model = _m.Texts(
            x=x, y=y, texts=text, size=size, color=color, family=family, anchor=anchor,
            rotation=rotation,
        )  # fmt: skip
        self.models.append(model)
        return model


class Axes(AxesBase, ModelsRef):
    """Layout model for 2D axes."""

    x: Axis = Field(default_factory=Axis, description="X-axis settings.")
    y: Axis = Field(default_factory=Axis, description="Y-axis settings.")
    axis_color: str = Field("#000000", description="Axis color.")


class SingleAxes(BaseLayoutModel):
    axes: Axes = Field(default_factory=Axes, description="Child axes.")

    @property
    def x(self) -> Axis:
        """X-axis settings."""
        return self.axes.x

    @property
    def y(self) -> Axis:
        """Y-axis settings."""
        return self.axes.y

    @property
    def axis_color(self) -> str:
        """Axis color."""
        return self.axes.axis_color

    @axis_color.setter
    def axis_color(self, value):
        self.axes.axis_color = value

    @property
    def title(self) -> str:
        """Title of the central axes."""
        return self.axes.title

    @title.setter
    def title(self, value):
        self.axes.title = value

    def merge_with(self, other: "SingleAxes") -> "SingleAxes":
        """Merge with another SingleAxes layout."""
        new_axes = self.axes.model_copy(
            update={"models": self.axes.models + other.axes.models}
        )
        return SingleAxes(axes=new_axes)

    ### Because there's only one axes, we can directly call the axes methods.
    def scatter(
        self,
        x: Sequence[float],
        y: Sequence[float] | None = None,
        *,
        symbol: str = "o",
        size: float | None = None,
        **kwargs,
    ) -> _m.Scatter:
        """Add a scatter plot model to the axes."""
        return self.axes.scatter(x=x, y=y, symbol=symbol, size=size, **kwargs)

    def plot(
        self,
        x: Sequence[float],
        y: Sequence[float] | None = None,
        **kwargs,
    ) -> _m.Line:
        """Add a line plot model to the axes."""
        return self.axes.plot(x=x, y=y, **kwargs)

    def bar(
        self,
        x: Sequence[float],
        y: Sequence[float] | None = None,
        *,
        bottom: "float | Sequence[float] | NDArray[np.number] | None" = None,
        bar_width: float | None = None,
        orient: Literal["vertical", "horizontal"] = "vertical",
        **kwargs,
    ) -> _m.Bar:
        """Add a bar plot model to the axes."""
        return self.axes.bar(
            x=x, y=y, bottom=bottom, bar_width=bar_width, orient=orient, **kwargs
        )

    def errorbar(
        self,
        x: Sequence[float],
        y: Sequence[float] | None = None,
        *,
        x_error: "float | Sequence[float] | NDArray[np.number] | None" = None,
        y_error: "float | Sequence[float] | NDArray[np.number] | None" = None,
        capsize: float | None = None,
        **kwargs,
    ) -> _m.ErrorBar:
        """Add an error bar plot model to the axes."""
        return self.axes.errorbar(
            x=x, y=y, x_error=x_error, y_error=y_error, capsize=capsize, **kwargs
        )

    def band(
        self,
        x: Sequence[float],
        y0: Sequence[float],
        y1: Sequence[float] | None = None,
        *,
        orient: Literal["vertical", "horizontal"] = "vertical",
        **kwargs,
    ) -> _m.Band:
        """Add a band plot model to the axes."""
        return self.axes.band(x=x, y0=y0, y1=y1, orient=orient, **kwargs)

    def hist(
        self,
        data: "Sequence[float] | NDArray[np.number]",
        *,
        bins: int = 10,
        range: tuple[float, float] | None = None,
        orient: Literal["vertical", "horizontal"] = "vertical",
        stat: Literal["count", "density", "probability"] = "count",
        **kwargs,
    ) -> _m.Histogram:
        """Add a histogram plot model to the axes."""
        return self.axes.hist(
            data=data, bins=bins, range=range, orient=orient, stat=stat, **kwargs
        )

    def text(
        self,
        x: Sequence[float],
        y: Sequence[float],
        text: Sequence[str],
        *,
        size: int = 12,
        color: str = "black",
        family: str = "Arial",
        rotation: float = 0,
        anchor: _m.ANCHOR_STRINGS = "center",
    ) -> _m.Texts:
        """Add a text plot model to the axes."""
        return self.axes.text(
            x=x, y=y, text=text, size=size, color=color, family=family, anchor=anchor,
            rotation=rotation,
        )  # fmt: skip


class StackedAxes(StackedAxesBase):
    """Layout model for stacked axes."""

    x: Axis = Field(default_factory=Axis, description="X-axis settings.")
    y: Axis = Field(default_factory=Axis, description="Y-axis settings.")
    axis_color: str = Field("#000000", description="Axis color.")

    def __getitem__(self, key) -> ModelsRef:
        if not isinstance(key, tuple):
            key = (key,)
        if len(key) != len(self.shape):
            raise ValueError(
                f"Key must be a tuple of length {len(self.shape)} but got: {key!r}"
            )
        if any(a >= b for a, b in zip(key, self.shape)):
            raise IndexError(
                f"Index {key!r} is out of bounds for shape {self.shape!r}."
            )
        if key not in self.models:
            models = self.models[key] = []
        else:
            models = self.models[key]
        return ModelsRef(models=models)


class SingleStackedAxes(BaseLayoutModel):
    model_type: ClassVar[str] = StandardType.PLOT_STACK

    axes: StackedAxes = Field(..., description="Child axes.")
    background_color: str = Field("#FFFFFF", description="Background color.")

    @property
    def x(self) -> Axis:
        """X-axis settings."""
        return self.axes.x

    @property
    def y(self) -> Axis:
        """Y-axis settings."""
        return self.axes.y

    @property
    def axis_color(self) -> str:
        """Axis color."""
        return self.axes.axis_color

    @axis_color.setter
    def axis_color(self, value):
        self.axes.axis_color = value

    @property
    def title(self) -> str:
        """Title of the central axes."""
        return self.axes.title

    @title.setter
    def title(self, value):
        self.axes.title = value

    @classmethod
    def fill(cls, *shape: int, multi_dims=None):
        models = {}
        for index in np.ndindex(*shape):
            models[index] = []
        return SingleStackedAxes(
            axes=StackedAxes(shape=shape, models=models, multi_dims=multi_dims)
        )

    @property
    def shape(self) -> tuple[int, ...]:
        """Shape of the stack dimensions."""
        return self.axes.shape

    def __getitem__(self, key) -> ModelsRef:
        return self.axes[key]


class Layout1D(BaseLayoutModel):
    """Layout model for 1D layout."""

    axes: list[Axes] = Field(default_factory=list, description="Child layouts.")
    share_x: bool = Field(False, description="Share x-axis or not.")
    share_y: bool = Field(False, description="Share y-axis or not.")

    def __getitem__(self, key: SupportsIndex) -> Axes:
        return self.axes[key]

    @classmethod
    def fill(cls, num: int):
        layout = cls()
        for _ in range(num):
            layout.axes.append(Axes())
        return layout

    def merge_with(self, other: "Self") -> "Self":
        if not isinstance(other, type(self)):
            raise ValueError(f"Cannot merge {type(self)} with {type(other)}")
        new_axes = [
            a.model_copy(update={"models": a.models + b.models})
            for a, b in zip(self.axes, other.axes)
        ]
        return type(self)(axes=new_axes, share_x=self.share_x, share_y=self.share_y)


class Row(Layout1D):
    """Layout model for row."""


class Column(Layout1D):
    """Layout model for column."""


class Grid(BaseLayoutModel):
    """Layout model for grid."""

    axes: list[list[Axes]] = Field(default_factory=list, description="Child layouts.")

    def __getitem__(self, key) -> Axes:
        return self.axes[key[0]][key[1]]

    @classmethod
    def fill(cls, rows: int, cols: int) -> "Self":
        layout = cls()
        for _ in range(rows):
            layout.axes.append([Axes() for _ in range(cols)])
        return layout

    def merge_with(self, other: "Self") -> "Self":
        if not isinstance(other, type(self)):
            raise ValueError(f"Cannot merge {type(self)} with {type(other)}")
        new_axes = [
            [
                a.model_copy(update={"models": a.models + b.models})
                for a, b in zip(row_a, row_b)
            ]
            for row_a, row_b in zip(self.axes, other.axes)
        ]
        return type(self)(axes=new_axes)


def _norm_xy(x, y) -> "tuple[NDArray[np.number], NDArray[np.number]]":
    if y is None:
        y = np.asarray(x)
        x = np.arange(len(y))
    else:
        x = np.asarray(x)
        y = np.asarray(y)
    if x.ndim != 1 or y.ndim != 1:
        raise ValueError(f"x and y must be 1D arrays, got {x.ndim}D and {y.ndim}D.")
    if x.shape != y.shape:
        raise ValueError(
            f"x and y must have the same shape, got {x.shape} and {y.shape}."
        )
    return x, y



from typing import Any, Callable, Literal, SupportsIndex
from himena import MainWindow
import numpy as np
from cmap import Colormap
from himena.data_wrappers._array import wrap_array
from himena.plugins import register_function, configure_gui, configure_submenu
from himena.types import Parametric, WidgetDataModel, Rect
from himena.consts import StandardType, MenuId
from himena.standards.model_meta import (
    DimAxis,
    ArrayMeta,
    ImageMeta,
    ImageRoisMeta,
    roi as _roi,
)
from himena.widgets import SubWindow, TabArea
from himena.utils import image_utils
from himena_builtins.tools.array import _cast_meta, _make_index_getter

configure_submenu(MenuId.TOOLS_IMAGE_ROI, "ROI")
configure_submenu(MenuId.TOOLS_IMAGE, group="20_builtins", order=6)
configure_submenu("/model_menu/roi", "ROI")


@register_function(
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_ROI, "/model_menu/roi"],
    command_id="builtins:image:crop-image",
    group="image-crop",
    keybindings=["Ctrl+Shift+X"],
)
def crop_image(model: WidgetDataModel) -> Parametric:
    """Crop the image around the current ROI."""
    arr = wrap_array(model.value)

    def _get_x(*_):
        bbox = _get_roi_box(model)
        return bbox.left, bbox.right

    def _get_y(*_):
        bbox = _get_roi_box(model)
        return bbox.top, bbox.bottom

    @configure_gui(x={"bind": _get_x}, y={"bind": _get_y})
    def run_crop_image(x: tuple[int, int], y: tuple[int, int]):
        xsl, ysl = slice(*x), slice(*y)
        meta = _cast_meta(model, ImageMeta)
        if meta.is_rgb:
            sl = (ysl, xsl, slice(None))
        else:
            sl = (ysl, xsl)
        arr_cropped = arr[(...,) + sl]
        meta_out = meta.without_rois()
        return model.with_value(
            arr_cropped.arr, metadata=meta_out
        ).with_title_numbering()

    return run_crop_image


@register_function(
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_ROI, "/model_menu/roi"],
    command_id="builtins:image:crop-image-multi",
    group="image-crop",
)
def crop_image_multi(model: WidgetDataModel) -> Parametric:
    """Crop the image around the registered ROIs and return as a model stack."""
    meta = _cast_meta(model, ImageMeta)
    arr = wrap_array(model.value)

    @configure_gui(bbox_list={"bind": _bbox_list_getter(meta, arr)})
    def run_crop_image_multi(bbox_list: list[tuple[int, int, int, int]]):
        meta_out = meta.without_rois()
        cropped_models: list[WidgetDataModel] = []
        for i, bbox in enumerate(bbox_list):
            sl = image_utils.bbox_to_slice(bbox, meta)
            arr_cropped = arr[(...,) + sl]
            model_0 = model.with_value(
                arr_cropped.arr, metadata=meta_out, title=f"ROI-{i} of {model.title}"
            )
            cropped_models.append(model_0)
        return WidgetDataModel(
            value=cropped_models,
            type=StandardType.MODELS,
            title=f"Cropped images from {model.title}",
        ).with_title_numbering()

    return run_crop_image_multi


def _bbox_list_getter(
    meta: ImageMeta,
    arr: np.ndarray,
) -> Callable[..., list[tuple[int, int, int, int]]]:
    def _get_bbox_list(*_):
        rois = meta.unwrap_rois()
        bbox_list: list[tuple[int, int, int, int]] = []
        for roi in rois:
            if not isinstance(roi, _roi.Roi2D):
                continue
            bbox = image_utils.roi_2d_to_bbox(roi, arr, meta.is_rgb)
            bbox_list.append(tuple(bbox))
        return bbox_list

    return _get_bbox_list


@register_function(
    title="Crop Image (nD) ...",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_ROI, "/model_menu/roi"],
    command_id="builtins:image:crop-image-nd",
    group="image-crop",
)
def crop_image_nd(win: SubWindow) -> Parametric:
    """Crop the image in nD, by drawing a 2D ROI and reading slider values."""
    from himena.qt.magicgui import SliderRangeGetter

    model = win.to_model()
    ndim = wrap_array(model.value).ndim
    meta = _cast_meta(model, ImageMeta)
    if ndim < 2:
        raise ValueError("Cannot crop image less than 2D.")
    axes_kwarg_names = [f"axis_{i}" for i in range(ndim)]
    image_axes = meta.axes
    if image_axes is None:
        image_axes = [DimAxis(name=f"axis-{i}") for i in range(ndim)]
    index_yx_rgb = 2 + int(meta.is_rgb)
    if ndim < index_yx_rgb + 1:
        raise ValueError("Image only has 2D data.")

    conf_kwargs = {}
    for i, axis_name in enumerate(axes_kwarg_names[:-index_yx_rgb]):
        conf_kwargs[axis_name] = {
            "widget_type": SliderRangeGetter,
            "getter": _make_index_getter(win, i),
            "label": image_axes[i].name,
        }
    axis_y, axis_x = axes_kwarg_names[ndim - index_yx_rgb : ndim - index_yx_rgb + 2]
    conf_kwargs[axis_y] = {"bind": _make_roi_limits_getter(win, "y")}
    conf_kwargs[axis_x] = {"bind": _make_roi_limits_getter(win, "x")}

    @configure_gui(**conf_kwargs)
    def run_crop_image(**kwargs: tuple[int | None, int | None]):
        model = win.to_model()  # NOTE: need to re-fetch the model
        arr = wrap_array(model.value)
        sl_nd = tuple(slice(x0, x1) for x0, x1 in kwargs.values())
        arr_cropped = arr[sl_nd]
        meta_out = meta.without_rois()
        meta_out.current_indices = None  # shape changed, need to reset
        return model.with_value(arr_cropped.arr, metadata=meta_out)

    return run_crop_image


def _make_roi_limits_getter(win: SubWindow, dim: Literal["x", "y"]):
    def _getter(*_):
        model = win.to_model()
        meta = _cast_meta(model, ImageMeta)
        roi = meta.current_roi
        arr = wrap_array(model.value)
        if roi is None:
            width, height = image_utils.slice_shape(arr, meta.is_rgb)
            left = top = 0
        else:
            if not isinstance(roi, _roi.Roi2D):
                raise NotImplementedError
            left, top, width, height = image_utils.roi_2d_to_bbox(roi, arr, meta.is_rgb)
        if dim == "x":
            return left, left + width
        return top, top + height

    return _getter


@register_function(
    title="Duplicate ROIs",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_ROI, "/model_menu/roi"],
    command_id="builtins:rois:duplicate",
)
def duplicate_rois(model: WidgetDataModel) -> WidgetDataModel:
    """Duplicate the ROIs as a new window with the ROI data."""
    # NOTE: this command will also be used by drag-and-drop of image view widget.
    if isinstance(meta := model.metadata, ArrayMeta):
        axes = meta.axes
    else:
        axes = None
    return WidgetDataModel(
        value=_get_rois_from_model(model),
        type=StandardType.ROIS,
        title=f"ROIs of {model.title}",
        metadata=ImageRoisMeta(axes=axes),
    )


@register_function(
    title="Filter ROIs",
    types=StandardType.ROIS,
    menus=["/model_menu"],
    command_id="builtins:rois:filter",
)
@register_function(
    title="Filter ROIs",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_ROI, "/model_menu/roi"],
    command_id="builtins:image:filter-rois",
)
def filter_rois(model: WidgetDataModel) -> Parametric:
    """Filter ROIs by their types."""
    rois = _get_rois_from_model(model)
    _choices = [
        "Rectangle", "RotatedRectangle", "Line", "SegmentedLine", "Point2D", "Points2D",
        "Ellipse", "Rotated Ellipse", "Polygon"
    ]  # fmt: skip

    @configure_gui(types={"choices": _choices, "widget_type": "Select"})
    def run_filter_rois(types: list[str]):
        types_allowed = {_roi.pick_roi_model(typ.lower()) for typ in types}
        sl = [i for i, r in enumerate(rois) if type(r) in types_allowed]
        value = rois.filter_by_selection(sl)
        axes = _axes_from_metadata(model)
        return WidgetDataModel(
            value=value,
            type=StandardType.ROIS,
            title=f"{model.title} filtered",
            metadata=ImageRoisMeta(axes=axes),
        )

    return run_filter_rois


@register_function(
    title="Select ROIs",
    types=StandardType.ROIS,
    menus=["/model_menu"],
    command_id="builtins:rois:select",
)
def select_rois(model: WidgetDataModel) -> Parametric:
    """Make a new ROI list with the selected ROIs."""
    rois = _get_rois_from_model(model)
    if not isinstance(meta := model.metadata, (ImageRoisMeta, ImageMeta)):
        raise ValueError(f"Expected an ImageRoisMeta metadata, got {type(meta)}")
    axes = meta.axes

    @configure_gui(selections={"bind": lambda *_: meta.selections})
    def run_select(selections: list[int]) -> WidgetDataModel:
        if len(selections) == 0:
            raise ValueError("No ROIs selected.")

        return WidgetDataModel(
            value=rois.filter_by_selection(selections),
            type=StandardType.ROIS,
            title=f"Subset of {model.title}",
            metadata=ImageRoisMeta(axes=axes),
        )

    return run_select


def _axes_from_metadata(model: WidgetDataModel) -> list[DimAxis] | None:
    if isinstance(meta := model.metadata, (ImageRoisMeta, ImageMeta)):
        axes = meta.axes
    else:
        axes = None
    return axes


@register_function(
    title="Point ROIs to DataFrame",
    types=StandardType.ROIS,
    menus=["/model_menu"],
    command_id="builtins:rois:point-rois-to-dataframe",
)
@register_function(
    title="Point ROIs to DataFrame",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_ROI, "/model_menu/roi"],
    command_id="builtins:image:point-rois-to-dataframe",
)
def point_rois_to_dataframe(model: WidgetDataModel) -> WidgetDataModel:
    """Convert point ROIs to a DataFrame."""
    rois = _get_rois_from_model(model)
    if len(rois) == 0:
        raise ValueError("No ROIs to convert")
    ndim = rois.indices.shape[1] + 2
    arrs: list[np.ndarray] = []
    for indices, roi in rois.iter_with_indices():
        if isinstance(roi, _roi.PointRoi2D):
            arr = np.array([indices + (roi.y, roi.x)], dtype=np.float32)
        elif isinstance(roi, _roi.PointsRoi2D):
            npoints = len(roi.xs)
            arr = np.empty((npoints, ndim))
            arr[:, :-2] = indices
            arr[:, -2] = roi.ys
            arr[:, -1] = roi.xs
        else:
            raise TypeError(f"Expected a PointRoi or PointsRoi, got {type(roi)}")
        arrs.append(arr)
    arr_all = np.concatenate(arrs, axis=0)
    axes = None
    if isinstance(meta := model.metadata, (ArrayMeta, ImageRoisMeta)):
        axes = meta.axes
    if axes is None:
        axes = [DimAxis(name=f"axis-{i}") for i in range(ndim)]
    out = {axis.name: arr_all[:, i] for i, axis in enumerate(axes)}
    return WidgetDataModel(
        value=out,
        type=StandardType.DATAFRAME,
        title=f"Points of {model.title}",
    )


@register_function(
    title="Set Colormap ...",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_CHANNELS, "/model_menu/channels"],
    command_id="builtins:image:set-colormaps",
    group="image-colormap",
)
def set_colormaps(model: WidgetDataModel) -> Parametric:
    """Set the colormaps for each channel."""
    from himena.qt.magicgui import ColormapEdit

    arr = wrap_array(model.value)
    meta = _cast_meta(model, ImageMeta).model_copy()
    if meta.channel_axis is None:
        channel_names = ["Channel"]
    else:
        nchn = arr.shape[meta.channel_axis]
        channel_names = _get_channel_names(meta, nchn, allow_single=True)
    current_channels = [ch.colormap for ch in meta.channels]
    colormap_defaults = [
        "gray", "green", "magenta", "cyan", "yellow", "red", "blue", "plasma",
        "viridis", "inferno", "imagej:fire", "imagej:HiLo", "imagej:ice", "matlab:jet",
        "matlab:hot",
    ]  # fmt: skip
    options = {
        f"ch_{i}": {
            "label": channel_names[i],
            "widget_type": ColormapEdit,
            "defaults": colormap_defaults,
            "value": current_channels[i],
        }
        for i in range(len(channel_names))
    }

    @configure_gui(gui_options=options, show_parameter_labels=len(channel_names) > 1)
    def set_cmaps(**kwargs: str):
        meta.channels = [
            ch.with_colormap(Colormap(cmap).name)
            for ch, cmap in zip(meta.channels, kwargs.values())
        ]
        return model.with_metadata(meta, update_inplace=True)

    return set_cmaps


@register_function(
    title="Propagate Colormaps",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_CHANNELS, "/model_menu/channels"],
    command_id="builtins:image:propagate-colormaps",
    group="image-colormap",
)
def propagate_colormaps(win: SubWindow, tab: TabArea):
    """Propagate the colormaps of current image to other images in this tab."""
    model = win.to_model()
    meta = _cast_meta(model, ImageMeta).model_copy()
    if meta.channel_axis is None:
        raise ValueError("Image does not have a channel axis.")
    if meta.is_rgb:
        raise ValueError("Image is RGB.")
    current_channels = meta.channels.copy()
    for each in tab:
        if each is win or each.model_type() != StandardType.IMAGE:
            continue
        each_model = each.to_model()
        if isinstance(each_meta := each_model.metadata, ImageMeta):
            if (
                each_meta.channel_axis is None
                or each_meta.is_rgb
                or len(each_meta.channels) != len(current_channels)
            ):
                continue
            each_meta.channels = [
                ch.with_colormap(current_channels[i].colormap)
                for i, ch in enumerate(each_meta.channels)
            ]

            each.update_model(each_model.with_metadata(each_meta))
    return None


@register_function(
    title="Split Channels",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_CHANNELS, "/model_menu/channels"],
    command_id="builtins:image:split-channels",
)
def split_channels(model: WidgetDataModel) -> list[WidgetDataModel]:
    """Split the image by the channel axis into separate images."""
    meta = _cast_meta(model, ImageMeta)
    arr = wrap_array(model.value)
    if meta.channel_axis is not None:
        c_axis = meta.channel_axis
        channel_labels = _get_channel_names(meta, arr.shape[c_axis])
    elif meta.is_rgb:
        c_axis = arr.ndim - 1
        channel_labels = ["R", "G", "B"]
    else:
        raise ValueError("Image does not have a channel axis and is not RGB.")
    slice_chn = (slice(None),) * c_axis
    models: list[WidgetDataModel] = []
    for idx in range(arr.shape[c_axis]):
        arr_i = arr[slice_chn + (idx,)]
        meta_i = meta.get_one_axis(c_axis, idx)
        meta_i.is_rgb = False
        title = f"[{channel_labels[idx]}] {model.title}"
        models.append(model.with_value(arr_i, metadata=meta_i, title=title))
    return models


@register_function(
    title="Set Channel Axis",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_CHANNELS, "/model_menu/channels"],
    command_id="builtins:image:set-channel-axis",
)
def set_channel_axis(model: WidgetDataModel) -> Parametric:
    """Set the channel axis of the image."""
    meta = _cast_meta(model, ImageMeta)
    if meta.channel_axis is not None:
        raise ValueError("Image already has a channel axis.")
    if meta.is_rgb:
        raise ValueError("Image is RGB.")
    arr = wrap_array(model.value)
    if meta.axes is None:
        choices = [(f"axis-{i}", i) for i in range(arr.ndim - 2)]
    else:
        choices = [(axis.name, i) for i, axis in enumerate(meta.axes[:-2])]

    @configure_gui(axis={"choices": choices})
    def run_set_channel_axis(axis: int, inplace: bool = False):
        meta_out = meta.model_copy(update={"channel_axis": axis})
        return model.with_metadata(meta_out, update_inplace=inplace)

    return run_set_channel_axis


@register_function(
    title="Specify Rectangle ...",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_ROI, "/model_menu/roi"],
    command_id="builtins:image:roi-specify-rectangle",
    group="image-spacify",
)
def roi_specify_rectangle(win: SubWindow) -> Parametric:
    """Specify the coordinates of a rectangle ROI."""

    model = win.to_model()
    meta = _cast_meta(model, ImageMeta)
    if isinstance(roi := meta.current_roi, _roi.RectangleRoi):
        x0, y0, w0, h0 = roi.x, roi.y, roi.width, roi.height
    else:
        nx, ny = image_utils.slice_shape(wrap_array(model.value), meta)
        x0, y0 = nx / 4, ny / 4
        w0, h0 = nx / 2, ny / 2

    @configure_gui(
        preview=True,
        x={"value": x0},
        y={"value": y0},
        width={"value": w0},
        height={"value": h0},
    )
    def run_roi_specify_coordinates(x: float, y: float, width: float, height: float):
        model = win.to_model()
        meta = _cast_meta(model, ImageMeta)
        indices = _slider_indices(meta)
        meta.current_roi = _roi.RectangleRoi(
            indices=indices, x=x, y=y, width=width, height=height
        )
        meta.skip_image_rerendering = True
        win.update_model(model.model_copy(update={"metadata": meta}))

    return run_roi_specify_coordinates


@register_function(
    title="Specify Ellipse ...",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_ROI, "/model_menu/roi"],
    command_id="builtins:image:roi-specify-ellipse",
    group="image-specify",
)
def roi_specify_ellipse(win: SubWindow) -> Parametric:
    """Specify the coordinates of an ellipse ROI."""

    model = win.to_model()
    meta = _cast_meta(model, ImageMeta)
    if isinstance(roi := meta.current_roi, _roi.EllipseRoi):
        x0, y0, w0, h0 = roi.x, roi.y, roi.width, roi.height
    else:
        nx, ny = image_utils.slice_shape(wrap_array(model.value), meta)
        x0, y0 = nx / 4, ny / 4
        w0, h0 = nx / 2, ny / 2

    @configure_gui(
        preview=True,
        x={"value": x0},
        y={"value": y0},
        width={"value": w0},
        height={"value": h0},
    )
    def run_roi_specify_coordinates(x: float, y: float, width: float, height: float):
        model = win.to_model()
        meta = _cast_meta(model, ImageMeta)
        indices = _slider_indices(meta)
        meta.current_roi = _roi.EllipseRoi(
            indices=indices, x=x, y=y, width=width, height=height
        )
        meta.skip_image_rerendering = True
        win.update_model(model.model_copy(update={"metadata": meta}))

    return run_roi_specify_coordinates


@register_function(
    title="Specify Line ...",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_ROI, "/model_menu/roi"],
    command_id="builtins:image:roi-specify-line",
    group="image-specify",
)
def roi_specify_line(win: SubWindow) -> Parametric:
    """Specify the coordinates of a line ROI."""

    model = win.to_model()
    meta = _cast_meta(model, ImageMeta)
    if isinstance(roi := meta.current_roi, _roi.LineRoi):
        x1, y1, x2, y2 = roi.x1, roi.y1, roi.x2, roi.y2
    else:
        nx, ny = image_utils.slice_shape(wrap_array(model.value), meta)
        x1, y1 = nx / 4, ny / 4
        x2, y2 = nx / 4 * 3, ny / 4 * 3

    @configure_gui(
        preview=True,
        x1={"value": x1},
        y1={"value": y1},
        x2={"value": x2},
        y2={"value": y2},
    )
    def run_roi_specify_coordinates(x1: float, y1: float, x2: float, y2: float):
        model = win.to_model()
        meta = _cast_meta(model, ImageMeta)
        indices = _slider_indices(meta)
        meta.current_roi = _roi.LineRoi(indices=indices, start=(x1, y1), end=(x2, y2))
        meta.skip_image_rerendering = True
        win.update_model(model.model_copy(update={"metadata": meta}))

    return run_roi_specify_coordinates


def _get_consensus_axes(models: list[WidgetDataModel]) -> list[str]:
    axes_consensus: list[str] | None = None
    for model in models:
        meta = _cast_meta(model, ImageMeta)
        if meta.axes is None:
            raise ValueError(f"Model {model!r} has no axes.")
        incoming_axes = [axis.name for axis in meta.axes]
        if axes_consensus is None:
            axes_consensus = incoming_axes
        elif axes_consensus != incoming_axes:
            raise ValueError("Images have different axes.")
    return axes_consensus


def _stack_and_insert_axis(
    arrs: list[Any],
    meta: ImageMeta,
    axis: str,
    axis_index: int,
    channel_axis: int | None = None,
) -> tuple[Any, ImageMeta]:
    arr_out = np.stack(arrs, axis=axis_index)
    if axes := meta.axes:
        axes = axes.copy()
        axes.insert(axis_index, DimAxis(name=axis))
    else:
        axes = None
    meta = meta.model_copy(
        update={
            "axes": axes,
            "channel_axis": channel_axis,
            "current_indices": None,
        }
    )
    return arr_out, meta


def run_merge_channels(
    images: list[SubWindow],
    axis: str = "c",
) -> WidgetDataModel:
    if len(images) < 2:
        raise ValueError("At least two images are required.")
    img_values = [img.to_model() for img in images]
    meta = _cast_meta(img_values[0], ImageMeta)
    if meta.is_rgb:
        raise ValueError("Cannot merge RGB image.")
    if meta.channel_axis is not None:
        raise ValueError("Image already has a channel axis.")
    arrs = [m.value for m in img_values]
    axes_consensus = _get_consensus_axes(img_values)
    c_axis = _find_index_to_insert_axis(axis, axes_consensus)
    axes_consensus.insert(c_axis, axis)

    arr_out, meta = _stack_and_insert_axis(
        arrs, meta, axis, c_axis, channel_axis=c_axis
    )
    # TODO: should colormaps be inherited?
    # colormaps = [_cast_meta(m, ImageMeta).channels[0].colormap for m in img_values]

    return img_values[0].with_value(
        arr_out, metadata=meta, title=f"[Merge] {img_values[0].title}"
    )


@register_function(
    title="Merge Channels ...",
    menus=[MenuId.TOOLS_IMAGE_CHANNELS, "/model_menu/channels"],
    command_id="builtins:image:merge-channels",
)
def merge_channels(ui: MainWindow) -> Parametric:
    """Stack images along the channel axis."""
    return configure_gui(run_merge_channels, images=_get_stack_options(ui))


@register_function(
    title="Stack Images ...",
    menus=[MenuId.TOOLS_IMAGE, "/model_menu/channels"],
    command_id="builtins:image:stack-images",
)
def stack_images(ui: MainWindow) -> Parametric:
    """Stack N-D images along a new axis to make a (N+1)-D image."""

    @configure_gui(images=_get_stack_options(ui))
    def run_stack_images(
        images: list[SubWindow],
        axis_name: str,
        axis_index: int | None = None,
    ) -> WidgetDataModel:
        """Stack images.

        Parameters
        ----------
        images : list[SubWindow]
            List of images to stack.
        axis_name : str
            Name of the new axis.
        axis_index : int, optional
            If provided, the index of the new axis in the stacked image. If not, new
            axis will be inserted to the most natural position.
        """
        if axis_name.title() in ("C", "Channel"):
            return run_merge_channels(images)
        if len(images) < 2:
            raise ValueError("At least two images are required.")
        img_values = [img.to_model() for img in images]
        if not isinstance(meta := img_values[0].metadata, ImageMeta):
            meta = ImageMeta()
        arrs = [m.value for m in img_values]
        axes_consensus = _get_consensus_axes(img_values)
        if axis_name == "":
            axis_name = _make_unique_axis_name(axes_consensus)
        if axis_index is None:
            axis_index = _find_index_to_insert_axis(axis_name, axes_consensus)
        axes_consensus.insert(axis_index, axis_name)
        if meta.channel_axis is not None:
            channel_axis = meta.channel_axis
            if axis_index <= channel_axis:
                channel_axis += 1
        else:
            channel_axis = None
        arr_out, meta = _stack_and_insert_axis(
            arrs, meta, axis_name, axis_index, channel_axis=channel_axis
        )
        return img_values[0].with_value(
            arr_out, metadata=meta, title=f"[Stack] {img_values[0].title}"
        )

    return run_stack_images


def _get_stack_options(ui: MainWindow) -> dict[str, Any]:
    if tab := ui.tabs.current():
        win_default = [win for win in tab if win.model_type() == StandardType.IMAGE]
        return {"types": [StandardType.IMAGE], "value": win_default}
    else:
        return {"types": [StandardType.IMAGE]}


def _get_current_roi_and_meta(
    model: WidgetDataModel,
) -> tuple[_roi.Roi2D, ImageMeta]:
    meta = _cast_meta(model, ImageMeta)
    if not (roi := meta.current_roi):
        raise ValueError("ROI selection is required for this operation.")
    return roi, meta


def _get_channel_names(
    meta: ImageMeta, nchannels: int, allow_single: bool = False
) -> list[str]:
    if meta.channel_axis is None:
        if allow_single:
            return ["Ch-0"]
        raise ValueError("Image does not have a channel axis.")
    if meta.axes is None:
        raise ValueError("Image does not have axes.")
    axis = meta.axes[meta.channel_axis]
    return [axis.get_label(i) for i in range(nchannels)]


def _make_unique_axis_name(axes: list[str]) -> str:
    i = 0
    while (axis := f"axis-{i}") in axes:
        i += 1
    return axis


def _find_index_to_insert_axis(axis: str, axes: list[str]) -> int:
    if axis in axes:
        raise ValueError(f"Axis '{axis}' already exists in axes: {axes!r}")
    _order_map = {"t": 0, "time": 0, "z": 1, "c": 2, "channel": 2}
    yx_axes_start_index = min(2, len(axes))
    # last axes are usually y, x
    axes_ref = [axis] + axes[:-yx_axes_start_index]
    axes_sorted = sorted(axes_ref, key=lambda x: _order_map.get(x.lower(), -1))
    return axes_sorted.index(axis)


def _slider_indices(meta: ImageMeta) -> tuple[int, ...]:
    if meta.current_indices:
        indices = tuple(i for i in meta.current_indices if isinstance(i, SupportsIndex))
    else:
        indices = ()
    return indices


def _get_rois_from_model(model: WidgetDataModel) -> _roi.RoiListModel:
    if model.type == StandardType.IMAGE:
        meta = _cast_meta(model, ImageMeta)
        rois = meta.unwrap_rois()
    elif model.type == StandardType.ROIS:
        rois = model.value
    else:
        raise ValueError(
            "Command can only be executed on an 'image' or 'image-rois' model."
        )
    return rois


def _get_roi_box(model: WidgetDataModel) -> Rect[int]:
    arr = wrap_array(model.value)
    roi, meta = _get_current_roi_and_meta(model)
    if not isinstance(roi, _roi.Roi2D):
        raise NotImplementedError
    return image_utils.roi_2d_to_bbox(roi, arr, meta.is_rgb)



from __future__ import annotations

from typing import TYPE_CHECKING
from himena.widgets import MainWindow
from himena.qt._qmain_window import QMainWindow

if TYPE_CHECKING:
    from app_model import Application
    from himena.style import Theme
    from qtpy import QtWidgets as QtW  # noqa: F401


class MainWindowQt(MainWindow["QtW.QWidget"]):
    """Main window with Qt backend."""

    _backend_main_window: QMainWindow

    def __init__(self, app: Application, theme: Theme) -> None:
        backend = QMainWindow(app=app)
        super().__init__(backend, app, theme)
        backend._himena_main_window = self
        backend._tab_widget._init_startup()
        backend._update_context()



from __future__ import annotations

from io import StringIO
import logging
from typing import TYPE_CHECKING, Any, cast
from dataclasses import dataclass
from qtpy import QtGui, QtCore, QtWidgets as QtW
from qtpy.QtCore import Qt
import numpy as np

from himena.data_wrappers import ArrayWrapper, wrap_array
from himena.consts import StandardType, MonospaceFontFamily
from himena.standards.model_meta import ArrayMeta
from himena.types import WidgetDataModel
from himena.plugins import validate_protocol
from himena.utils.collections import UndoRedoStack
from himena.utils.misc import is_structured
from himena_builtins.qt.widgets._table_components import (
    QTableBase,
    Editability,
    QSelectionRangeEdit,
    format_table_value,
    FLAGS,
    parse_string,
)

if TYPE_CHECKING:
    from himena_builtins.qt.widgets._table_components import SelectionModel


_LOGGER = logging.getLogger(__name__)


class QArrayModel(QtCore.QAbstractTableModel):
    """Table model for data frame."""

    def __init__(self, arr: np.ndarray, parent: QArrayView | None = None):
        super().__init__(parent)
        self._arr_slice = arr  # 2D
        self._slice: tuple[int, ...] = ()
        if not is_structured(arr):
            if arr.ndim != 2:
                raise ValueError("Only 2D array is supported.")
            self._dtype = np.dtype(arr.dtype)
            self._nrows, self._ncols = arr.shape
            self._get_dtype = self._get_dtype_nonstructured
            self._get_item = self._get_item_nonstructured
        else:
            if arr.ndim != 1 or not is_structured(arr):
                raise ValueError(
                    f"Only 1D structured array is supported (got {arr.ndim}D array "
                    f"with dtype {arr.dtype!r})."
                )
            self._dtype = arr.dtype
            self._nrows, self._ncols = arr.shape[0], len(arr.dtype.names)
            self._get_dtype = self._get_dtype_structured
            self._get_item = self._get_item_structured
        self._is_original_array = True

    def _get_dtype_nonstructured(self, r: int, c: int) -> np.dtype:
        return self._dtype

    def _get_dtype_structured(self, r: int, c: int) -> np.dtype:
        return self._dtype.fields[self._dtype.names[c]][0]

    def _get_item_nonstructured(self, r: int, c: int) -> Any:
        return self._arr_slice[r, c]

    def _get_item_structured(self, r: int, c: int) -> Any:
        return self._arr_slice[r][self._dtype.names[c]]

    def flags(self, index: QtCore.QModelIndex) -> Qt.ItemFlag:
        return FLAGS

    def rowCount(self, parent=None):
        return self._nrows

    def columnCount(self, parent=None):
        return self._ncols

    def data(
        self,
        index: QtCore.QModelIndex,
        role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole,
    ):
        if not index.isValid():
            return QtCore.QVariant()
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            r, c = index.row(), index.column()
            if r < self.rowCount() and c < self.columnCount():
                if self._get_dtype(r, c).kind in "iuf":
                    return Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter
                else:
                    return Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter
        elif role == Qt.ItemDataRole.ToolTipRole:
            r, c = index.row(), index.column()
            array_indices = ", ".join(str(i) for i in self._slice + (r, c))
            return f"A[{array_indices}] = {self._get_item(r, c)!r}"
        elif role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            r, c = index.row(), index.column()
            if r < self.rowCount() and c < self.columnCount():
                value = self._get_item(r, c)
                if role == Qt.ItemDataRole.DisplayRole:
                    text = format_table_value(value, self._get_dtype(r, c).kind)
                else:
                    text = str(value)
                return text
        return QtCore.QVariant()

    def setData(self, index: QtCore.QModelIndex, value: Any, role: int = ...) -> bool:
        if role == Qt.ItemDataRole.EditRole:
            self.parent()._table.set_string_input(index.row(), index.column(), value)
            return True
        return False

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ):
        if role == Qt.ItemDataRole.DisplayRole:
            if (
                is_structured(self._arr_slice)
                and orientation == Qt.Orientation.Horizontal
            ):
                return self._dtype.names[section]
            return str(section)
        elif role == Qt.ItemDataRole.ToolTipRole:
            if (
                is_structured(self._arr_slice)
                and orientation == Qt.Orientation.Horizontal
            ):
                name = self._dtype.names[section]
                return f"{name} (dtype: {self._dtype.fields[name][0]})"
            return None

    if TYPE_CHECKING:

        def parent(self) -> QArrayView: ...


class QArraySliceView(QTableBase):
    def __init__(self, parent):
        super().__init__(parent)
        self.horizontalHeader().setFixedHeight(18)
        self.verticalHeader().setDefaultSectionSize(20)
        self.horizontalHeader().setDefaultSectionSize(55)
        self.setModel(QArrayModel(np.zeros((0, 0))))
        self.setFont(QtGui.QFont(MonospaceFontFamily, 10))

    def update_width_by_dtype(self):
        kind = self.model()._dtype.kind
        depth = self.model()._dtype.itemsize
        if kind in "ui":
            self._update_width(min(depth * 40, 55))
        elif kind == "f":
            self._update_width(min(depth * 40, 55))
        elif kind == "c":
            self._update_width(min(depth * 40 + 8, 55))
        else:
            self._update_width(55)

    def _update_width(self, width: int):
        header = self.horizontalHeader()
        header.setDefaultSectionSize(width)
        for i in range(header.count()):
            header.resizeSection(i, width)

    def set_array(self, arr: np.ndarray, slice_):
        self.model()._arr_slice = arr
        self.model()._slice = slice_
        self.update()

    def set_string_input(self, r: int, c: int, value: str):
        view = self.parent()
        sl = view._get_indices() + (r, c)
        view.array_update(sl, value, record_undo=True)

    def _copy_data(self):
        sels = self._selection_model.ranges
        if len(sels) > 1:
            _LOGGER.warning("Multiple selections.")
            return
        sel = sels[0]
        arr_slice = self.model()._arr_slice
        buf = StringIO()
        if is_structured(arr_slice):
            fields = [
                arr_slice.dtype.names[i] for i in range(sel[1].start, sel[1].stop)
            ]
            arr_to_copy = arr_slice[sel[0]][fields]
        else:
            arr_to_copy = arr_slice[sel]
        fmt = dtype_to_fmt(self.model()._dtype)
        np.savetxt(
            buf,
            arr_to_copy,
            delimiter="\t",
            fmt=fmt,
        )
        clipboard = QtW.QApplication.clipboard()
        clipboard.setText(buf.getvalue().rstrip("\n"))

    def _paste_from_clipboard(self):
        text = QtW.QApplication.clipboard().text()
        if not text:
            return

        buf = StringIO(text)
        arr_paste = np.loadtxt(
            buf, dtype=np.dtypes.StringDType(), delimiter="\t", ndmin=2
        )
        # undo info
        rng = self._selection_model.get_single_range()
        if rng[0].stop - rng[0].start == 1 and rng[1].stop - rng[1].start == 1:
            # 1x1 selection can be pasted to any size
            rng = (
                slice(rng[0].start, rng[0].start + arr_paste.shape[0]),
                slice(rng[1].start, rng[1].start + arr_paste.shape[1]),
            )
        sl = self.parent()._get_indices() + rng
        _ud_old_data = self.parent()._arr.get_slice(sl).copy()

        self._paste_array(sl, arr_paste)

        # undo info
        _ud_new_data = arr_paste.copy()
        _action_edit = EditAction(_ud_old_data, _ud_new_data, sl)
        self.parent()._undo_stack.push(_action_edit)

        # select what was just pasted
        self._selection_model.set_ranges([rng])
        self.update()

    def _paste_array(self, sl, arr_paste: np.ndarray):
        arr_slice = self.model()._arr_slice
        arr = self.parent()._arr

        rng = self._selection_model.get_single_range()
        row0, col0 = rng[0].start, rng[1].start
        lr = max(arr_paste.shape[0], rng[0].stop - rng[0].start)
        lc = max(arr_paste.shape[1], rng[1].stop - rng[1].start)

        # expand the table if necessary
        if (row0 + lr) > arr_slice.shape[0] or (col0 + lc) > arr_slice.shape[1]:
            raise ValueError("Cannot paste outside of the array.")

        # paste the data
        if is_structured(arr):
            raise NotImplementedError("Structured array paste is not implemented.")
        else:
            arr[sl] = arr_paste
        self.parent()._spinbox_changed()

    def _make_context_menu(self):
        menu = QtW.QMenu(self)
        menu.addAction("Copy", self._copy_data)
        menu.addAction("Paste", self._paste_from_clipboard)
        return menu

    if TYPE_CHECKING:

        def model(self) -> QArrayModel: ...
        def parent(self) -> QArrayView: ...


class QArrayView(QtW.QWidget):
    """A widget for viewing n-D arrays.

    ## Basic Usage

    The 2D array sliced for the last dimensions (such as A[2, 1, :, :]) are shown in the
    table. "2D array" can be any numpy-like arrays, including `xarray.DataArray`,
    `dask.array.Array`, `cupy.ndarray`, etc. If the array is more than 2D, spinboxes are
    shown in the bottom of the widget to select the slice. Numpy structured arrays are
    also supported.

    ## Copying Data

    Selected range can be copied `Ctrl+C`. The copied data is in tab-separated format so
    that it can be pasted to spreadsheet softwares.

    ## Editing Data

    You can edit the data in the current slice of the array. Input text will be parsed
    to the dtype of the array.
    """

    __himena_widget_id__ = "builtins:QArrayView"
    __himena_display_name__ = "Bulit-in Array Viewer"

    def __init__(self):
        super().__init__()
        self._table = QArraySliceView(self)
        layout = QtW.QVBoxLayout(self)

        self._spinbox_group = QtW.QWidget()
        group_layout = QtW.QHBoxLayout(self._spinbox_group)
        group_layout.setContentsMargins(1, 1, 1, 1)
        group_layout.setAlignment(Qt.AlignmentFlag.AlignRight)
        group_layout.addWidget(QtW.QLabel("Slice:"))

        layout.addWidget(self._table)
        layout.addWidget(self._spinbox_group)
        self._spinboxes: list[QtW.QSpinBox] = []
        self._arr: ArrayWrapper | None = None
        self._control: QArrayViewControl | None = None
        self._model_type = StandardType.ARRAY
        self._undo_stack = UndoRedoStack[EditAction](size=20)
        self._axes = None

    @property
    def selection_model(self) -> SelectionModel:
        """The selection model of the array slice view."""
        return self._table.selection_model

    def set_indices(self, *indices):
        for sb, idx in zip(self._spinboxes, indices, strict=True):
            sb.setValue(idx)
        self._spinbox_changed()

    def _update_spinbox_for_shape(self, shape: tuple[int, ...], dims_shown: int = 2):
        nspin = len(self._spinboxes)
        # make insufficient spinboxes
        for _i in range(nspin, len(shape) - dims_shown):
            self._make_spinbox(shape[_i])

        for i, sb in enumerate(self._spinboxes):
            if i < len(shape) - dims_shown:
                sb.setVisible(True)
                _max = shape[i] - 1
                if sb.value() > _max:
                    sb.setValue(_max)
                sb.setRange(0, _max)
            else:
                self._spinbox_group.layout().removeWidget(sb)
                sb.deleteLater()
                self._spinboxes.remove(sb)

        self._spinbox_group.setVisible(len(shape) > dims_shown)

    def _spinbox_changed(self):
        arr = self._arr
        if arr is None:
            return
        sl = self._get_slice()
        arr = self._arr.get_slice(sl)
        if arr.ndim < 2 and not is_structured(self._arr.arr):
            arr = arr.reshape(-1, 1)
        self._table.set_array(arr, sl)

    def _get_slice(self) -> tuple[int | slice, ...]:
        if self._arr.ndim < 2:
            return (slice(None),)
        arr_structured = is_structured(self._arr.arr)
        if arr_structured:
            last_sl = (slice(None),)
        else:
            last_sl = (slice(None), slice(None))
        return self._get_indices() + last_sl

    def _get_indices(self) -> tuple[int, ...]:
        return tuple(sb.value() for sb in self._spinboxes)

    def _make_spinbox(self, max_value: int):
        spinbox = QtW.QSpinBox()
        self._spinbox_group.layout().addWidget(spinbox)
        spinbox.setRange(0, max_value - 1)
        spinbox.valueChanged.connect(self._spinbox_changed)
        self._spinboxes.append(spinbox)

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        was_none = self._arr is None
        arr = wrap_array(model.value)
        self._arr = arr
        arr_structured = is_structured(arr.arr)
        self._update_spinbox_for_shape(arr.shape, dims_shown=1 if arr_structured else 2)
        if arr.ndim < 2:
            arr_slice = arr.get_slice(())
            if is_structured(arr_slice):
                self._table.setModel(QArrayModel(arr_slice, self))
            else:
                self._table.setModel(QArrayModel(arr_slice.reshape(-1, 1), self))
        else:
            sl = self._get_slice()
            self._table.setModel(QArrayModel(arr.get_slice(sl), self))

        if self._control is None:
            self._control = QArrayViewControl()
        self._control.update_for_array(self)
        if was_none:
            self._table.update_width_by_dtype()
        if isinstance(meta := model.metadata, ArrayMeta):
            if meta.selections:  # if has selections, they need updating
                self.selection_model.clear()
            for (r0, r1), (c0, c1) in meta.selections:
                self.selection_model.append((slice(r0, r1), slice(c0, c1)))
            self._axes = meta.axes

        self._model_type = model.type
        self._modified_override = None
        self.update()
        return None

    @validate_protocol
    def to_model(self) -> WidgetDataModel[list[list[Any]]]:
        # NOTE: if this model is passed to another widget and modified in this widget,
        # the other dataframe will be modified as well. To avoid this, we need to reset
        # the copy-on-write state of this array.
        self._table.model()._is_original_array = True
        current_indices = tuple(
            None if isinstance(sl, slice) else sl for sl in self._get_slice()
        )
        return WidgetDataModel(
            value=self._arr.arr,
            type=self.model_type(),
            extension_default=".npy",
            metadata=ArrayMeta(
                axes=self._axes,
                current_indices=current_indices,
                selections=self._table._get_selections(),
            ),
        )

    @validate_protocol
    def model_type(self) -> str:
        return self._model_type

    @validate_protocol
    def size_hint(self):
        return 320, 280

    @validate_protocol
    def is_modified(self) -> bool:
        if self._modified_override is not None:
            return self._modified_override
        return self._undo_stack.undoable()

    @validate_protocol
    def set_modified(self, value: bool) -> None:
        self._modified_override = value

    @validate_protocol
    def is_editable(self) -> bool:
        return self._table.editTriggers() == Editability.TRUE

    @validate_protocol
    def set_editable(self, value: bool) -> None:
        if value:
            trig = Editability.TRUE
        else:
            trig = Editability.FALSE
        self._table.setEditTriggers(trig)

    @validate_protocol
    def control_widget(self):
        return self._control

    def array_update(
        self,
        sl,
        value: Any,
        *,
        record_undo: bool = True,
    ) -> None:
        """Update the data at the given index."""
        _ud_old_data = self._arr[sl]
        if isinstance(_ud_old_data, ArrayWrapper):
            _ud_old_data = _ud_old_data.copy()
        if self._table.model()._is_original_array:
            self._arr = self._arr.copy()
            self._table.model()._is_original_array = False
            self._spinbox_changed()  # update slice
        if isinstance(value, str):
            self._arr[sl] = parse_string(value, self._arr.dtype.kind)
        else:
            # called in undo/redo
            self._arr[sl] = value
        _ud_new_data = self._arr[sl]
        _action = EditAction(_ud_old_data, _ud_new_data, sl)
        if record_undo:
            self._undo_stack.push(_action)

    def undo(self):
        if action := self._undo_stack.undo():
            action.invert().apply(self)
            self.update()

    def redo(self):
        if action := self._undo_stack.redo():
            action.apply(self)
            self.update()

    def keyPressEvent(self, e: QtGui.QKeyEvent) -> None:
        _Ctrl = QtCore.Qt.KeyboardModifier.ControlModifier
        _mod = e.modifiers()
        _key = e.key()
        if _mod & _Ctrl and _key == QtCore.Qt.Key.Key_Z:
            self.undo()
            return
        elif _mod & _Ctrl and _key == QtCore.Qt.Key.Key_Y:
            self.redo()
            return
        if _mod & _Ctrl and _key == QtCore.Qt.Key.Key_C:
            return self._table._copy_data()
        if _mod & _Ctrl and _key == QtCore.Qt.Key.Key_V:
            return self._table._paste_from_clipboard()
        if _mod & _Ctrl and _key == QtCore.Qt.Key.Key_F:
            self._table._find_string()
            return
        return super().keyPressEvent(e)

    def setFocus(self) -> None:
        self._table.setFocus()


_R_CENTER = Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter


class QArrayViewControl(QtW.QWidget):
    def __init__(self):
        super().__init__()
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(_R_CENTER)
        self._label = QtW.QLabel("")
        self._label.setAlignment(_R_CENTER)
        layout.addWidget(self._label)
        self._selection_range = QSelectionRangeEdit()
        layout.addWidget(self._selection_range)

    def update_for_array(self, widget: QArrayView):
        if widget is None:
            return None
        _type_desc = widget.model_type()
        arr = widget._arr
        if arr is None:
            self._label.setText("No data available.")
            return None
        if not is_structured(arr):
            self._label.setText(f"{_type_desc} {arr.shape!r} {arr.dtype}")
        else:
            ncols = len(arr.dtype.names)
            self._label.setText(f"{_type_desc} {arr.shape!r} x {ncols} fields")
        self._selection_range.connect_table(widget._table)
        return None


def dtype_to_fmt(dtype: np.dtype) -> str:
    """Choose a proper format string for the dtype to convert to text."""
    if dtype.kind == "fc":
        dtype = cast(np.number, dtype)
        s = 1 if dtype.kind == "f" else 2
        if dtype.itemsize / s == 2:
            # 16bit has 10bit (~10^3) fraction
            return "%.4e"
        if dtype.itemsize / s == 4:
            # 32bit has 23bit (~10^7) fraction
            return "%.8e"
        if dtype.itemsize / s == 8:
            # 64bit has 52bit (~10^15) fraction
            return "%.16e"
        if dtype.itemsize / s == 16:
            # 128bit has 112bit (~10^33) fraction
            return "%.34e"
        raise RuntimeError(f"Unsupported float dtype: {dtype}")

    if dtype.kind in "iub":
        return "%d"
    return "%s"


@dataclass
class EditAction:
    old: Any
    new: Any
    index: Any

    def invert(self) -> EditAction:
        return EditAction(self.new, self.old, self.index)

    def apply(self, table: QArrayView):
        return table.array_update(self.index, self.new, record_undo=False)



from __future__ import annotations

from cmap import Colormap
import numpy as np
from numpy.testing import assert_allclose
import pytest
from himena import WidgetDataModel, create_image_model
from himena.standards.model_meta import ImageChannel, ImageMeta, DimAxis
from himena.standards import roi as _roi
from himena.testing.subwindow import WidgetTester


def test_setting_colormap(tester: WidgetTester):
    tester.update_model(_zyx_image_model())
    model = tester.to_model()
    meta = _cast_meta(model.metadata)
    assert len(meta.channels) == 1
    assert Colormap(meta.channels[0].colormap) == Colormap("gray")

    tester.update_model(_zyx_image_model(colormap="green"))
    model = tester.to_model()
    meta = _cast_meta(model.metadata)
    if len(meta.channels) != 1:
        raise AssertionError(f"Expected 1 channel, got {len(meta.channels)}")
    if Colormap(meta.channels[0].colormap) != Colormap("green"):
        raise AssertionError(
            f"Expected {Colormap('green')}, got {meta.channels[0].colormap}"
        )


def test_setting_unit(tester: WidgetTester):
    tester.update_model(_zyx_image_model())
    model = tester.to_model()
    meta = _cast_meta(model.metadata)
    if meta.unit != "a.u.":
        raise AssertionError(f"Expected 'a.u.', got {meta.unit}")

    tester.update_model(_zyx_image_model(unit="mV"))
    model = tester.to_model()
    meta = _cast_meta(model.metadata)
    if meta.unit != "mV":
        raise AssertionError(f"Expected 'mV', got {meta.unit}")


def test_setting_pixel_scale(tester: WidgetTester):
    tester.update_model(_zyx_image_model())
    model = tester.to_model()
    meta = _cast_meta(model.metadata)
    assert len(meta.axes) == 3
    if any(axis.scale != pytest.approx(1.0) for axis in meta.axes):
        raise AssertionError(
            f"Expected scales [1.0, 1.0, 1.0], got {[axis.scale for axis in meta.axes]}"
        )

    scales = [0.1, 0.2, 0.3]
    tester.update_model(_zyx_image_model(pixel_scale=scales))
    model = tester.to_model()
    meta = _cast_meta(model.metadata)
    assert meta.axes is not None
    assert len(meta.axes) == 3
    if any(
        axis.scale != pytest.approx(scale) for axis, scale in zip(meta.axes, scales)
    ):
        raise AssertionError(
            f"Expected scales {scales}, got {[axis.scale for axis in meta.axes]}"
        )


def test_setting_axis_names(tester: WidgetTester):
    tester.update_model(_zyx_image_model())
    model = tester.to_model()
    meta = _cast_meta(model.metadata)
    assert len(meta.axes) == 3
    assert meta.axes is not None
    if any(axis.name != name for axis, name in zip(meta.axes, ["z", "y", "x"])):
        raise AssertionError(
            f"Expected names ['z', 'y', 'x'], got {[axis.name for axis in meta.axes]}"
        )

    names = ["t", "y", "x"]
    tester.update_model(_zyx_image_model(axis_names=names))
    model = tester.to_model()
    meta = _cast_meta(model.metadata)
    assert meta.axes is not None
    assert len(meta.axes) == 3
    if any(axis.name != name for axis, name in zip(meta.axes, names)):
        raise AssertionError(
            f"Expected names {names}, got {[axis.name for axis in meta.axes]}"
        )


def test_setting_current_indices(tester: WidgetTester):
    tester.update_model(_zyx_image_model())
    model = tester.to_model()
    meta = _cast_meta(model.metadata)
    if meta.current_indices != (0, None, None):
        raise AssertionError(f"Expected (0, None, None), got {meta.current_indices}")

    tester.update_model(_zyx_image_model(current_indices=(1, None, None)))
    model = tester.to_model()
    meta = _cast_meta(model.metadata)
    if meta.current_indices != (1, None, None):
        raise AssertionError(f"Expected (1, None, None), got {meta.current_indices}")


def test_current_roi(tester: WidgetTester):
    tester.update_model(_zyx_image_model())
    model = tester.to_model()
    meta = _cast_meta(model.metadata)
    assert meta.current_roi is None

    roi = _roi.RectangleRoi(indices=(1,), x=2.9, y=0, width=2, height=2.5)
    tester.update_model(
        _zyx_image_model(current_roi=roi, current_indices=(1, None, None))
    )
    model = tester.to_model()
    meta = _cast_meta(model.metadata)
    assert meta.current_roi is not None
    croi = meta.current_roi
    assert isinstance(croi, _roi.RectangleRoi)
    assert_allclose([croi.x, croi.y, croi.width, croi.height], [2.9, 0, 2, 2.5])


def test_change_dimensionality(tester: WidgetTester):
    """Check changing the dimensionality of the image works."""
    tester.update_model(create_image_model(np.arange(15).reshape(3, 5)))
    assert tester.to_model().value.shape == (3, 5)
    assert not tester.to_model().metadata.is_rgb
    tester.update_model(create_image_model(np.arange(24).reshape(4, 6)))
    assert tester.to_model().value.shape == (4, 6)
    assert not tester.to_model().metadata.is_rgb
    tester.update_model(create_image_model(np.arange(96).reshape(4, 4, 6)))
    assert tester.to_model().value.shape == (4, 4, 6)
    assert not tester.to_model().metadata.is_rgb


def test_rgb_images(tester: WidgetTester):
    """Check that RGB images are handled correctly."""
    tester.update_model(
        create_image_model(np.full((5, 5, 3), 200, dtype=np.uint8), is_rgb=True),
    )
    assert tester.to_model().value.shape == (5, 5, 3)
    assert tester.to_model().metadata.is_rgb
    tester.update_model(
        create_image_model(np.full((4, 6, 3), 200, dtype=np.uint8), is_rgb=True)
    )
    assert tester.to_model().value.shape == (4, 6, 3)
    assert tester.to_model().metadata.is_rgb
    tester.update_model(create_image_model(np.full((4, 6), 100, dtype=np.uint8)))
    assert tester.to_model().value.shape == (4, 6)
    assert not tester.to_model().metadata.is_rgb


def test_current_roi_and_its_index(tester: WidgetTester):
    """Check that current ROI and its index are handled correctly."""
    roi0 = _roi.RectangleRoi(x=2.9, y=0, width=2, height=2.5)
    roi1 = _roi.RectangleRoi(x=1.5, y=0, width=2, height=2.5)
    roi2 = _roi.RectangleRoi(x=0.5, y=0, width=2, height=2.5)
    rois = _roi.RoiListModel(items=[roi0, roi1, roi2])
    for i in range(3):
        tester.update_model(
            create_image_model(
                np.arange(15).reshape(3, 5),
                current_roi=rois[i],
                current_roi_index=i,
                rois=rois,
            )
        )
        cur_roi = tester.to_model().metadata.current_roi
        assert cur_roi is not None
        assert tester.to_model().metadata.current_roi_index == i
        assert cur_roi.x == rois[i].x
        assert cur_roi.y == rois[i].y
        assert cur_roi.width == rois[i].width
        assert cur_roi.height == rois[i].height


def _cast_meta(meta) -> ImageMeta:
    assert isinstance(meta, ImageMeta)
    return meta


def _zyx_image_model(
    axis_names: list[str] = ["z", "y", "x"],
    pixel_scale: list[float] = [1.0, 1.0, 1.0],
    pixel_unit: str = "um",
    colormap: str = "gray",
    unit: str = "a.u.",
    current_indices=(0, None, None),
    current_roi: _roi.RoiModel | None = None,
) -> WidgetDataModel:
    axes = [
        DimAxis(name=name, scale=scale, unit=pixel_unit)
        for name, scale in zip(axis_names, pixel_scale)
    ]
    channels = [ImageChannel(colormap=colormap)]
    return create_image_model(
        np.arange(48).reshape(3, 4, 4),
        axes=axes,
        unit=unit,
        channels=channels,
        current_indices=current_indices,
        current_roi=current_roi,
        is_rgb=False,
    )



"""Builtin File explorer plugin."""

import sys
from dataclasses import dataclass
from himena.plugins import (
    register_dock_widget_action,
    add_default_status_tip,
    config_field,
)

add_default_status_tip(
    short="File Explorer",
    long="Ctrl+Shift+E to open a file explorer widget on the left.",
)


@dataclass
class FileExplorerConfig:
    allow_drop_data_to_save: bool = config_field(
        default=True,
        tooltip="Allow dropping data opened in the main window to save.",
    )
    allow_drop_file_to_move: bool = config_field(
        default=True, tooltip="Allow dropping files to move."
    )


@dataclass
class FileExplorerSSHConfig:
    default_host: str = config_field(
        default="", tooltip="The default host name or IP address"
    )
    default_user: str = config_field(default="", tooltip="The default user name")
    default_port: int = config_field(default=22, tooltip="The default port number")
    default_use_wsl: bool = config_field(
        default=False,
        tooltip="Use WSL to connect to the host in Windows",
        enabled=sys.platform == "win32",
    )
    default_protocol: str = config_field(
        default="rsync",
        tooltip="The default protocol to use (rsync or scp)",
    )


@register_dock_widget_action(
    title="File Explorer",
    area="left",
    keybindings="Ctrl+Shift+E",
    command_id="builtins:file-explorer",
    singleton=True,
    plugin_configs=FileExplorerConfig(),
)
def make_file_explorer_widget(ui):
    """Open a file explorer widget as a dock widget."""
    from himena_builtins.qt.explorer._widget import QExplorerWidget

    return QExplorerWidget(ui)


@register_dock_widget_action(
    title="Remote File Explorer",
    area="left",
    command_id="builtins:file-explorer-ssh",
    singleton=True,
    plugin_configs=FileExplorerSSHConfig(),
)
def make_file_explorer_ssh_widget(ui):
    """Open a remote file explorer widget as a dock widget."""
    from himena_builtins.qt.explorer._widget_ssh import QSSHRemoteExplorerWidget

    return QSSHRemoteExplorerWidget(ui)



from himena.consts import StandardType
from himena.plugins import register_widget_class
from himena_builtins.qt.widgets.array import QArrayView

register_widget_class(StandardType.ARRAY, QArrayView, priority=50)



import warnings
import numpy as np
from numpy.testing import assert_equal
from pathlib import Path
import pytest
from qtpy import QtCore
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QApplication
from himena.standards.model_meta import DimAxis, ImageMeta
from pytestqt.qtbot import QtBot
from himena import MainWindow, StandardType, create_image_model
from himena.standards.roi import RoiListModel, LineRoi, PointRoi2D, PointsRoi2D
from himena.standards.roi.core import RectangleRoi
from himena.testing import WidgetTester, image, file_dialog_response
from himena.types import WidgetDataModel
from himena.widgets import SubWindow
from himena_builtins.qt.widgets.image import QImageView, QImageLabelView
from himena_builtins.qt.widgets._image_components import _roi_items as _rois
from himena_builtins.qt.widgets._image_components._control import ComplexMode
from himena_builtins.tools.image import _make_roi_limits_getter, _bbox_list_getter
from himena_builtins.qt.widgets._image_components._handles import RoiSelectionHandles

_Ctrl = Qt.KeyboardModifier.ControlModifier

def test_image_view(qtbot: QtBot):
    image_view = QImageView()
    image_view.show()
    image_view._img_view.set_show_labels(True)
    with WidgetTester(image_view) as tester:
        # grayscale
        tester.update_model(value=np.arange(100, dtype=np.uint8).reshape(10, 10))
        qtbot.addWidget(image_view)
        assert len(image_view._dims_slider._sliders) == 0

        # 5D with channel
        rng = np.random.default_rng(14442)
        tester.update_model(
            value=rng.random((10, 5, 3, 100, 100), dtype=np.float32),
            metadata=ImageMeta(channel_axis=2)
        )
        image_view._dims_slider._sliders[0]._slider.setValue(1)
        image_view._dims_slider._sliders[2]._slider.setValue(2)
        image_view._control._chn_mode_combo.setCurrentText("Gray")
        QApplication.processEvents()
        image_view._control._chn_mode_combo.setCurrentText("Mono")
        QApplication.processEvents()
        image_view._control._chn_mode_combo.setCurrentText("Comp.")
        QApplication.processEvents()
        image_view._control._histogram._line_low._show_value_label()

        # switch modes
        _shift = Qt.KeyboardModifier.ShiftModifier
        _view = image_view._img_view
        qtbot.keyClick(_view, Qt.Key.Key_Z)
        assert _view._mode == _view.Mode.PAN_ZOOM
        qtbot.keyClick(_view, Qt.Key.Key_L)
        assert _view._mode == _view.Mode.ROI_LINE
        qtbot.keyPress(_view, Qt.Key.Key_Space)
        assert _view._mode == _view.Mode.PAN_ZOOM
        qtbot.keyRelease(_view, Qt.Key.Key_Space)
        assert _view._mode == _view.Mode.ROI_LINE
        qtbot.keyClick(_view, Qt.Key.Key_L, modifier=_shift)
        assert _view._mode == _view.Mode.ROI_SEGMENTED_LINE
        qtbot.keyClick(_view, Qt.Key.Key_P)
        assert _view._mode == _view.Mode.ROI_POINT
        qtbot.keyClick(_view, Qt.Key.Key_P, modifier=_shift)
        assert _view._mode == _view.Mode.ROI_POINTS
        qtbot.keyClick(_view, Qt.Key.Key_R)
        assert _view._mode == _view.Mode.ROI_RECTANGLE
        qtbot.keyClick(_view, Qt.Key.Key_R, modifier=_shift)
        assert _view._mode == _view.Mode.ROI_ROTATED_RECTANGLE
        qtbot.keyClick(_view, Qt.Key.Key_E)
        assert _view._mode == _view.Mode.ROI_ELLIPSE
        qtbot.keyClick(_view, Qt.Key.Key_E, modifier=_shift)
        assert _view._mode == _view.Mode.ROI_ROTATED_ELLIPSE
        qtbot.keyClick(_view, Qt.Key.Key_C)
        assert _view._mode == _view.Mode.ROI_CIRCLE
        qtbot.keyClick(_view, Qt.Key.Key_G)
        assert _view._mode == _view.Mode.ROI_POLYGON

        # click sliders
        slider = image_view._dims_slider._sliders[0]
        assert not slider._edit_value_line.isVisible()
        qtbot.mouseDClick(slider._index_label, Qt.MouseButton.LeftButton)
        assert slider._edit_value_line.isVisible()
        qtbot.keyClick(slider._edit_value_line, Qt.Key.Key_Escape)
        assert not slider._edit_value_line.isVisible()
        qtbot.mouseDClick(slider._index_label, Qt.MouseButton.LeftButton)
        assert slider._edit_value_line.isVisible()
        slider._edit_value_line.setText("2")
        qtbot.keyClick(slider._edit_value_line, Qt.Key.Key_Return)
        assert not slider._edit_value_line.isVisible()
        assert slider._slider.value() == 2

        image_view._img_view._make_menu_for_view()

def test_image_labels_view(qtbot: QtBot):
    image_view = QImageLabelView()
    image_view.show()
    with WidgetTester(image_view) as tester:
        tester.update_model(value=np.arange(24, dtype=np.uint8).reshape(2, 3, 4))
        qtbot.addWidget(image_view)
        assert len(image_view._dims_slider._sliders) == 1

        image_view._dims_slider._sliders[0]._slider.setValue(1)

def test_image_view_rgb(qtbot: QtBot):
    image_view = QImageView()
    image_view.show()
    with WidgetTester(image_view) as tester:
        tester.update_model(
            value=np.zeros((100, 100, 3), dtype=np.uint8), metadata=ImageMeta(is_rgb=True),
        )
        qtbot.addWidget(image_view)
        tester.cycle_model()
        assert len(image_view._dims_slider._sliders) == 0
        image_view._control._interp_check_box.setChecked(False)
        image_view._control._interp_check_box.setChecked(True)
        image_view._control._chn_mode_combo.setCurrentText("Gray")
        QApplication.processEvents()
        image_view._control._chn_mode_combo.setCurrentText("Color")
        QApplication.processEvents()

        tester.update_model(
            value=np.zeros((100, 100, 4), dtype=np.uint8), metadata=ImageMeta(is_rgb=True),
        )
        tester.cycle_model()
        assert len(image_view._dims_slider._sliders) == 0
        image_view._control._interp_check_box.setChecked(False)
        image_view._control._interp_check_box.setChecked(True)
        image_view._control._chn_mode_combo.setCurrentText("Gray")
        QApplication.processEvents()
        image_view._control._chn_mode_combo.setCurrentText("Color")
        QApplication.processEvents()

@pytest.mark.parametrize("unit", ["", "nm"])
def test_image_view_draw_roi(qtbot: QtBot, unit: str):
    image_view = QImageView()
    image_view.show()
    with WidgetTester(image_view) as tester:
        tester.update_model(
            value=np.zeros((100, 100), dtype=np.uint8),
            metadata=ImageMeta(
                axes=[DimAxis(name="y", unit=unit), DimAxis(name="x", unit=unit)]
            ),
        )
        qtbot.addWidget(image_view)

        # test ROI drawing
        vp = image_view._img_view.viewport()
        # pan zoom
        image_view._img_view.switch_mode(image_view._img_view.Mode.PAN_ZOOM)
        qtbot.mousePress(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(10, 10))
        qtbot.mouseMove(vp, pos=QtCore.QPoint(50, 50))
        qtbot.mouseRelease(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))
        qtbot.mouseDClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))

        # rectangle
        image_view._img_view.switch_mode(image_view._img_view.Mode.ROI_RECTANGLE)
        qtbot.mousePress(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(10, 10))
        qtbot.mouseMove(vp, pos=QtCore.QPoint(50, 50))
        qtbot.mouseRelease(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))
        assert isinstance(image_view._img_view._current_roi_item, _rois.QRectangleRoi)
        qtbot.keyPress(image_view._img_view, Qt.Key.Key_T)
        assert len(image_view._img_view._roi_items) == 1
        image_view._img_view._make_menu_for_roi(image_view._img_view._roi_items[0])
        qtbot.mouseDClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))

        # ellipse
        image_view._img_view.switch_mode(image_view._img_view.Mode.ROI_ELLIPSE)
        qtbot.mousePress(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(20, 20))
        qtbot.mouseMove(vp, pos=QtCore.QPoint(50, 50))
        qtbot.mouseRelease(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))
        assert isinstance(image_view._img_view._current_roi_item, _rois.QEllipseRoi)
        qtbot.mouseDClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))

        # line
        image_view._img_view.switch_mode(image_view._img_view.Mode.ROI_LINE)
        qtbot.mousePress(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(10, 10))
        qtbot.mouseMove(vp, pos=QtCore.QPoint(50, 50))
        qtbot.mouseRelease(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))
        assert isinstance(image_view._img_view._current_roi_item, _rois.QLineRoi)
        qtbot.mouseDClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(30, 10))
        # line should be removed by clicking somewhere else
        assert image_view._img_view._current_roi_item is None

        # rotated rect
        image_view._img_view.switch_mode(image_view._img_view.Mode.ROI_ROTATED_RECTANGLE)
        qtbot.mousePress(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(10, 10))
        qtbot.mouseMove(vp, pos=QtCore.QPoint(50, 50))
        qtbot.mouseRelease(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))
        assert isinstance(image_view._img_view._current_roi_item, _rois.QRotatedRectangleRoi)
        qtbot.mouseDClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(80, 80))
        # roi should be removed by clicking somewhere else
        assert image_view._img_view._current_roi_item is None

        # rotated ellipse
        image_view._img_view.switch_mode(image_view._img_view.Mode.ROI_ROTATED_ELLIPSE)
        qtbot.mousePress(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(10, 10))
        qtbot.mouseMove(vp, pos=QtCore.QPoint(50, 50))
        qtbot.mouseRelease(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))
        assert isinstance(image_view._img_view._current_roi_item, _rois.QRotatedEllipseRoi)
        qtbot.mouseDClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(80, 80))
        # roi should be removed by clicking somewhere else
        assert image_view._img_view._current_roi_item is None

        # circle
        image_view._img_view.switch_mode(image_view._img_view.Mode.ROI_CIRCLE)
        qtbot.mousePress(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(10, 10))
        qtbot.mouseMove(vp, pos=QtCore.QPoint(50, 50))
        qtbot.mouseRelease(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))
        assert isinstance(image_view._img_view._current_roi_item, _rois.QCircleRoi)
        qtbot.mouseDClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 60))
        # circle should be removed by clicking somewhere else
        assert image_view._img_view._current_roi_item is None

        # polygon
        image_view._img_view.switch_mode(image_view._img_view.Mode.ROI_POLYGON)
        qtbot.mouseClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(10, 10))
        qtbot.mouseClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(10, 20))
        qtbot.mouseClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(20, 20))
        qtbot.mouseClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(20, 30))
        qtbot.mouseDClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(30, 30))
        assert isinstance(image_view._img_view._current_roi_item, _rois.QPolygonRoi)
        qtbot.keyClick(image_view._img_view, Qt.Key.Key_Delete)
        qtbot.mouseDClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))
        assert image_view._img_view._current_roi_item is None

        # segmented line
        image_view._img_view.switch_mode(image_view._img_view.Mode.ROI_SEGMENTED_LINE)
        qtbot.mousePress(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(10, 10))
        qtbot.mouseMove(vp, pos=QtCore.QPoint(20, 20))
        qtbot.mouseRelease(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(20, 20))
        qtbot.mousePress(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(20, 20))
        qtbot.mouseMove(vp, pos=QtCore.QPoint(30, 20))
        qtbot.mouseRelease(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(30, 20))
        qtbot.mousePress(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(30, 20))
        qtbot.mouseRelease(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(30, 20))
        qtbot.mouseDClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))

        # point
        image_view._img_view.switch_mode(image_view._img_view.Mode.ROI_POINT)
        qtbot.mousePress(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(10, 10))
        qtbot.mouseMove(vp, pos=QtCore.QPoint(20, 20))
        qtbot.mouseRelease(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(20, 20))
        qtbot.mousePress(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(40, 40))
        qtbot.mouseRelease(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(40, 40))
        assert isinstance(image_view._img_view._current_roi_item, _rois.QPointRoi)
        qtbot.mouseDClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))


        # select
        image_view._img_view.switch_mode(image_view._img_view.Mode.SELECT)
        qtbot.mousePress(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(20, 20))
        qtbot.mouseMove(vp, pos=QtCore.QPoint(50, 50))
        qtbot.mouseRelease(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))
        qtbot.mouseDClick(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))


def test_image_view_copy_roi(himena_ui: MainWindow, qtbot: QtBot):
    image_view = QImageView()
    himena_ui.add_widget(image_view)
    himena_ui.show()
    with WidgetTester(image_view) as tester:
        tester.update_model(value=np.zeros((100, 100), dtype=np.uint8))
        qtbot.addWidget(image_view)

        # draw rectangle
        vp = image_view._img_view.viewport()
        image_view._img_view.switch_mode(image_view._img_view.Mode.ROI_RECTANGLE)
        qtbot.mousePress(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(10, 10))
        qtbot.mouseMove(vp, pos=QtCore.QPoint(50, 50))
        qtbot.mouseRelease(vp, Qt.MouseButton.LeftButton, pos=QtCore.QPoint(50, 50))

        qtbot.keyClick(image_view, Qt.Key.Key_C, modifier=_Ctrl)
        image_view._img_view.standard_ctrl_key_press(Qt.Key.Key_V)
        assert len(image_view._img_view._roi_items) == 2
        image_view._img_view.standard_ctrl_key_press(Qt.Key.Key_V)
        assert len(image_view._img_view._roi_items) == 3

def test_image_view_copy_roi_from_window_to_window(himena_ui: MainWindow, qtbot: QtBot):
    image_view_0 = QImageView()
    image_view_1 = QImageView()
    himena_ui.add_widget(image_view_0)
    himena_ui.add_widget(image_view_1)
    himena_ui.show()
    with (
        WidgetTester(image_view_0) as tester_0,
        WidgetTester(image_view_1) as tester_1
    ):
        tester_0.update_model(
            value=np.zeros((100, 100), dtype=np.uint8),
            metadata=ImageMeta(
                rois=RoiListModel(
                    items=[LineRoi(name="ROI-0", start=(1, 1), end=(4, 5),)],
                )
            )
        )
        tester_1.update_model(value=np.zeros((100, 100), dtype=np.uint8))
        qtbot.addWidget(image_view_0)
        qtbot.addWidget(image_view_1)
        image_view_0._roi_col._list_view.set_selection([0])
        qtbot.keyClick(image_view_0, Qt.Key.Key_C, modifier=_Ctrl)
        image_view_1._img_view.standard_ctrl_key_press(Qt.Key.Key_V)
        assert len(image_view_0._img_view._roi_items) == 1
        assert len(image_view_1._img_view._roi_items) == 1
        qtbot.keyClick(image_view_1._roi_col._list_view, Qt.Key.Key_V, modifier=_Ctrl)
        assert len(image_view_0._img_view._roi_items) == 1
        assert len(image_view_1._img_view._roi_items) == 2


def test_image_view_select_roi(qtbot: QtBot):
    image_view = QImageView()
    image_view.resize(150, 150)
    image_view.show()
    image_view.setSizes([300, 100])
    with WidgetTester(image_view) as tester:
        tester.update_model(value=np.zeros((100, 100), dtype=np.uint8))
        qtbot.addWidget(image_view)
        view = image_view._img_view
        view._wheel_event(1)
        view._wheel_event(-1)
        # point
        view._current_roi_item = _rois.QPointRoi(2, 3)
        view.select_item_at(QtCore.QPointF(2, 3))
        assert isinstance(view._current_roi_item, _rois.QPointRoi)
        view.select_item_at(QtCore.QPointF(10, 10))
        assert view._current_roi_item is None

        # points
        view._current_roi_item = _rois.QPointsRoi([2, 4], [3, 4])
        view.select_item_at(QtCore.QPointF(2, 3))
        assert isinstance(view._current_roi_item, _rois.QPointsRoi)
        view.select_item_at(QtCore.QPointF(10, 10))
        assert view._current_roi_item is None

        # line
        view._current_roi_item = _rois.QLineRoi(0, 0, 3, 3)
        view.select_item_at(QtCore.QPointF(1, 1))
        assert isinstance(view._current_roi_item, _rois.QLineRoi)
        view.select_item_at(QtCore.QPointF(3, 0))
        assert view._current_roi_item is None

        # rectangle
        view._current_roi_item = _rois.QRectangleRoi(0, 0, 3, 3)
        view.select_item_at(QtCore.QPointF(1, 2))
        assert isinstance(view._current_roi_item, _rois.QRectangleRoi)
        view.select_item_at(QtCore.QPointF(10, 2))
        assert view._current_roi_item is None

        # ellipse
        view._current_roi_item = _rois.QEllipseRoi(0, 0, 3, 5)
        view.select_item_at(QtCore.QPointF(1, 2))
        assert isinstance(view._current_roi_item, _rois.QEllipseRoi)
        view.select_item_at(QtCore.QPointF(0, 4))
        # assert view._current_roi_item is None  # FIXME: Not working for some reason

        # polygon
        view._current_roi_item = _rois.QPolygonRoi([0, 1, 3, 0], [3, 5, 5, 3])
        view.select_item_at(QtCore.QPointF(1, 5))
        assert isinstance(view._current_roi_item, _rois.QPolygonRoi)
        view.select_item_at(QtCore.QPointF(6, 3))
        assert view._current_roi_item is None

        # segmented line
        view._current_roi_item = _rois.QSegmentedLineRoi([0, 1, 3], [3, 5, 5])
        view.select_item_at(QtCore.QPointF(1, 5))
        assert isinstance(view._current_roi_item, _rois.QSegmentedLineRoi)
        view.select_item_at(QtCore.QPointF(6, 3))
        assert view._current_roi_item is None

        # rotated rectangle
        view._current_roi_item = _rois.QRotatedRectangleRoi(
            QtCore.QPointF(0, 0),
            QtCore.QPointF(10, 10),
            6,
        )
        view.select_item_at(QtCore.QPointF(4, 4))
        assert isinstance(view._current_roi_item, _rois.QRotatedRectangleRoi)
        view.select_item_at(QtCore.QPointF(10, 0))
        assert view._current_roi_item is None

        # rotated ellipse
        view._current_roi_item = _rois.QRotatedEllipseRoi(
            QtCore.QPointF(0, 0),
            QtCore.QPointF(10, 10),
            6,
        )
        view.select_item_at(QtCore.QPointF(4, 4))
        assert isinstance(view._current_roi_item, _rois.QRotatedEllipseRoi)
        view.select_item_at(QtCore.QPointF(10, 0))
        assert view._current_roi_item is None

        # circle
        view._current_roi_item = _rois.QCircleRoi(0, 0, 5)
        view.select_item_at(QtCore.QPointF(1, 2))
        assert isinstance(view._current_roi_item, _rois.QCircleRoi)
        view.select_item_at(QtCore.QPointF(1, 5))
        assert view._current_roi_item is None

        image_view._roi_col._roi_labels_btn.click()
        QApplication.processEvents()
        assert image_view._roi_col._roi_visible_btn.isChecked()
        assert not image_view._roi_col._roi_labels_btn.isChecked()
        image_view._roi_col._roi_visible_btn.click()
        QApplication.processEvents()
        assert not image_view._roi_col._roi_visible_btn.isChecked()
        assert not image_view._roi_col._roi_labels_btn.isChecked()
        image_view._roi_col._roi_labels_btn.click()
        QApplication.processEvents()
        assert image_view._roi_col._roi_visible_btn.isChecked()
        assert image_view._roi_col._roi_labels_btn.isChecked()
        image_view._roi_col._list_view._prep_context_menu(
            image_view._roi_col._list_view.model().index(0, 0)
        )
        qtbot.mouseClick(image_view._roi_col._list_view.viewport(), Qt.MouseButton.LeftButton)
        qtbot.mouseMove(image_view._roi_col._list_view.viewport(), QtCore.QPoint(3, 3))
        qtbot.mouseMove(image_view._roi_col._list_view.viewport(), QtCore.QPoint(4, 4))
        image_view._roi_col.remove_selected_rois()
        QApplication.processEvents()
        image_view._img_view.set_current_roi(_rois.QPolygonRoi([0, 1, 3, 0], [3, 5, 5, 3]))
        QApplication.processEvents()
        image_view._img_view.remove_current_item()

def test_image_view_roi_selection_from_list_widget(qtbot: QtBot):
    """Check if clicking ROIs from the QRoiCollection works as expected."""
    image_view = QImageView()
    image_view.setSizes([300, 100])
    with WidgetTester(image_view) as tester:
        tester.update_model(
            value=np.zeros((100, 100), dtype=np.uint8),
            metadata=ImageMeta(
                rois=RoiListModel(
                    items=[
                        LineRoi(name="ROI-0", start=(1, 1), end=(4, 5)),
                        PointRoi2D(name="ROI-1", x=1, y=5),
                    ],
                ),
            ),
        )
        qtbot.addWidget(image_view)
        assert image_view._roi_col.count() == 2
        list_view = image_view._roi_col._list_view
        image_view._roi_col._on_item_clicked(list_view.model().index(0, 0))
        assert isinstance(image_view._img_view._current_roi_item, _rois.QLineRoi)
        image_view._roi_col._on_item_clicked(list_view.model().index(1, 0))
        assert isinstance(image_view._img_view._current_roi_item, _rois.QPointRoi)
        assert len(image_view._img_view._roi_items) == 2
        assert any(isinstance(item, _rois.QLineRoi) for item in image_view._img_view.items())
        assert any(isinstance(item, _rois.QPointRoi) for item in image_view._img_view.items())

def test_image_view_roi_actions(qtbot: QtBot):
    image_view = QImageView()
    image_view.setSizes([300, 100])
    with WidgetTester(image_view) as tester:
        tester.update_model(
            value=np.zeros((5, 2, 30, 30), dtype=np.uint8),
            metadata=ImageMeta(
                axes=[DimAxis(name=name) for name in ["t", "c", "y", "x"]],
                rois=RoiListModel(
                    items=[
                        LineRoi(name="ROI-0", start=(1, 1), end=(4, 5)),
                        PointRoi2D(name="ROI-1", x=1, y=5),
                        RectangleRoi(name="ROI-2", x=1, y=1, width=3, height=4),
                    ],
                    indices=np.array([[0, 0], [1, 0], [0, 1]], dtype=np.int32),
                    axis_names=["t", "c"],
                ),
            ),
        )
        qtbot.addWidget(image_view)
        assert image_view._roi_col.count() == 3
        image_view._roi_col.flatten_roi(0)
        image_view._roi_col.flatten_roi_along(1, axis=1)
        # check context menu on flattened ROI
        list_view = image_view._roi_col._list_view

        list_view._prep_context_menu(list_view.model().index(0, 0))
        list_view._prep_context_menu(list_view.model().index(1, 0))
        image_view._roi_col.move_roi(2, (3, 0))
        meta = tester.to_model().metadata
        assert isinstance(meta, ImageMeta)
        assert_equal(meta.unwrap_rois().indices, [[-1, -1], [1, -1], [3, 0]])
        opt = image_view._roi_col._list_view._options_for_select_dimensions()
        dlg, container = image_view._roi_col._list_view._make_dialog(opt)
        qtbot.addWidget(dlg)
        assert len(container) == 2
        assert container[0].label == "t"
        assert container[1].label == "c"

        idx0 = list_view.model().index(0, 0)
        idx_invalid = list_view.model().index(10, 0)
        for idx in [idx0, idx_invalid]:
            list_view.model().data(idx, Qt.ItemDataRole.DisplayRole)
            list_view.model().data(idx, Qt.ItemDataRole.EditRole)
            list_view.model().data(idx, Qt.ItemDataRole.DecorationRole)
            list_view.model().data(idx, Qt.ItemDataRole.FontRole)
            list_view.model().data(idx, Qt.ItemDataRole.SizeHintRole)
            list_view.model().data(idx, Qt.ItemDataRole.ToolTipRole)
        list_view.model().setData(idx0, "new name", Qt.ItemDataRole.EditRole)
        assert list_view.indexAt(QtCore.QPoint(10, 10)).isValid()
        list_view._hover_event(QtCore.QPoint(10, 10))
        assert not list_view.indexAt(QtCore.QPoint(10, 1000)).isValid()
        list_view._hover_event(QtCore.QPoint(10, 1000))

def test_constrast_hist(qtbot: QtBot):
    image_view = QImageView()
    image_view.show()
    with WidgetTester(image_view) as tester:
        tester.update_model(value=np.zeros((100, 100), dtype=np.uint8))
        qtbot.addWidget(image_view)
        control = image_view.control_widget()
        qtbot.addWidget(control)
        control._auto_cont_btn.click()
        control._histogram.set_clim((1, 2))
        control._histogram._make_context_menu()
        control._histogram._set_hist_scale_func("linear")
        control._histogram._set_hist_scale_func("log")
        control._histogram._img_to_clipboard()
        control._histogram._reset_view()

def test_complex_image(qtbot: QtBot):
    image_view = QImageView()
    image_view.show()
    yy, xx = np.indices((5, 5))
    img = np.exp(-1j * (yy + xx))
    with WidgetTester(image_view) as tester, warnings.catch_warnings():
        warnings.simplefilter("error")
        tester.update_model(value=img)
        qtbot.addWidget(image_view)
        control = image_view.control_widget()
        qtbot.addWidget(control)
        control.show()
        assert control._cmp_mode_combo.isVisible()
        control._cmp_mode_combo.setCurrentText(ComplexMode.REAL)
        control._cmp_mode_combo.setCurrentText(ComplexMode.IMAG)
        control._cmp_mode_combo.setCurrentText(ComplexMode.ABS)
        control._cmp_mode_combo.setCurrentText(ComplexMode.LOG_ABS)
        control._cmp_mode_combo.setCurrentText(ComplexMode.PHASE)

def test_image_view_change_dimensionality(qtbot: QtBot):
    image.test_change_dimensionality(_get_tester())

def test_image_view_setting_colormap(qtbot: QtBot):
    image.test_setting_colormap(_get_tester())

def test_image_view_setting_unit(qtbot: QtBot):
    image.test_setting_unit(_get_tester())

def test_image_view_setting_axis_names(qtbot: QtBot):
    image.test_setting_axis_names(_get_tester())

def test_image_view_setting_pixel_scale(qtbot: QtBot):
    image.test_setting_pixel_scale(_get_tester())

def test_image_view_setting_current_indices(qtbot: QtBot):
    image.test_setting_current_indices(_get_tester())

def test_image_view_current_roi(qtbot: QtBot):
    image.test_current_roi(_get_tester())

def test_image_view_current_roi_index(qtbot: QtBot):
    image.test_current_roi_and_its_index(_get_tester())

def _get_tester():
    return WidgetTester(QImageView())

def test_crop_image(himena_ui: MainWindow, tmpdir):
    model = create_image_model(
        np.zeros((4, 4, 10, 10)),
        axes=["t", "z", "y", "x"],
        current_roi=RectangleRoi(indices=(0, 0), x=1, y=1, width=6, height=4),
    )
    win = himena_ui.add_data_model(model)
    himena_ui.exec_action("builtins:image:crop-image")
    himena_ui.exec_action("builtins:image:copy-slice-to-clipboard")
    with file_dialog_response(himena_ui, Path(tmpdir) / "tmp.png"):
        himena_ui.exec_action("builtins:image:save-slice")
    himena_ui.current_window = win
    himena_ui.exec_action(
        "builtins:image:crop-image-multi",
        with_params={"bbox_list": [(1, 1, 4, 5), (1, 5, 2, 2)]}
    )
    himena_ui.current_window = win
    himena_ui.exec_action(
        "builtins:array:crop-nd",
        with_params={"axis_0": (2, 4), "axis_1": (0, 1), "axis_2": (1, 5), "axis_3": (2, 8)},
    )

    # multi-channel image
    model = create_image_model(
        np.zeros((4, 3, 10, 10)),
        axes=["t", "c", "y", "x"],
        current_roi=RectangleRoi(indices=(0, 0), x=1, y=1, width=6, height=4),
        channel_axis=1,
    )
    win = himena_ui.add_data_model(model)
    himena_ui.exec_action(
        "builtins:image:capture-setting",
        with_params={
            "scale_bars": [
                {"shape": (10, 1.5), "color": "red", "anchor_pos": "top-left", "offset": (2, 2)}
            ],
        }
    )

    himena_ui.exec_action("builtins:image:crop-image")
    himena_ui.exec_action("builtins:image:copy-slice-to-clipboard")
    with file_dialog_response(himena_ui, Path(tmpdir) / "tmp.png"):
        himena_ui.exec_action("builtins:image:save-slice")
    himena_ui.current_window = win
    himena_ui.exec_action(
        "builtins:image:crop-image-multi",
        with_params={"bbox_list": [(1, 1, 4, 5), (1, 5, 2, 2)]}
    )
    himena_ui.current_window = win
    himena_ui.exec_action(
        "builtins:array:crop-nd",
        with_params={"axis_0": (2, 4), "axis_1": (0, 1), "axis_2": (1, 5), "axis_3": (2, 8)},
    )

    # test limit getters
    model = win.to_model()
    _bbox_list_getter(model.metadata, model.value)()
    _make_roi_limits_getter(win, "x")()
    _make_roi_limits_getter(win, "y")()

def test_image_view_commands(himena_ui: MainWindow, tmpdir):
    himena_ui.add_data_model(
        WidgetDataModel(
            value=np.zeros((20, 20)),
            type=StandardType.IMAGE,
        )
    )
    himena_ui.exec_action("builtins:image:set-zoom-factor", with_params={"scale": 100.0})
    himena_ui.exec_action("builtins:image:copy-viewer-screenshot")
    with file_dialog_response(himena_ui, Path(tmpdir) / "tmp.png") as save_path:
        himena_ui.exec_action("builtins:image:save-viewer-screenshot")
        assert save_path.exists()


def test_roi_commands(himena_ui: MainWindow):
    model = create_image_model(
        np.zeros((4, 4, 10, 10)),
        axes=["t", "z", "y", "x"],
        rois=RoiListModel(
            items=[
                LineRoi(name="ROI-0", start=(1, 1), end=(4, 5)),
                PointRoi2D(name="ROI-1", x=1, y=5),
            ],
            indices=np.array([[0, 0], [0, 0]], dtype=np.int32),
            axis_names=["t", "z"],
        ),
    )
    win = himena_ui.add_data_model(model)
    himena_ui.exec_action("builtins:rois:duplicate")
    assert isinstance(lmodel := himena_ui.current_model.value, RoiListModel)
    assert len(lmodel) == 2
    win_roi = himena_ui.current_window
    himena_ui.exec_action(
        "builtins:rois:filter",
        with_params={"types": ["Line"]},
    )
    himena_ui.current_window = win_roi
    himena_ui.exec_action(
        "builtins:rois:select",
        with_params={"selections": [1]},
    )

    # specify
    himena_ui.current_window = win
    himena_ui.exec_action(
        "builtins:image:roi-specify-rectangle",
        with_params={"x": 3, "y": 2, "width": 3.0, "height": 3.0}
    )
    himena_ui.exec_action(
        "builtins:image:roi-specify-ellipse",
        with_params={"x": 3, "y": 2, "width": 3.0, "height": 3.0}
    )
    himena_ui.exec_action(
        "builtins:image:roi-specify-line",
        with_params={"x1": 3, "y1": 2, "x2": 3.0, "y2": 3.0}
    )

    himena_ui.add_object(
        RoiListModel(
            [PointRoi2D(x=0, y=0), PointsRoi2D(xs=[2, 3], ys=[1, 2])],
        ),
        type=StandardType.ROIS,
    )

    himena_ui.exec_action("builtins:image:point-rois-to-dataframe")
    assert himena_ui.current_model.type == StandardType.DATAFRAME

    # colormap
    model = create_image_model(
        np.zeros((4, 2, 10, 10)),
        axes=["t", "c", "y", "x"],
        channel_axis=1,
    )

    win = himena_ui.add_data_model(model)
    himena_ui.exec_action("builtins:image:set-colormaps", with_params={"ch_0": "green", "ch_1": "red"})
    assert isinstance(meta := win.to_model().metadata, ImageMeta)
    assert len(meta.channels) == 2
    assert meta.channels[0].colormap == "cmap:green"
    assert meta.channels[1].colormap == "cmap:red"
    himena_ui.exec_action("builtins:image:split-channels")
    win_g = himena_ui.tabs.current()[-2]
    win_r = himena_ui.tabs.current()[-1]
    assert win_g.to_model().metadata.colormap == "cmap:green"
    assert win_r.to_model().metadata.colormap == "cmap:red"
    himena_ui.exec_action("builtins:image:merge-channels", with_params={"images": [win_g, win_r]})
    himena_ui.exec_action(
        "builtins:image:stack-images",
        with_params={"images": [win_g, win_r], "axis_name": "p"}
    )

def test_rgb(himena_ui: MainWindow):
    model = create_image_model(
        np.zeros((10, 10, 3)),
        axes=["y", "x", "c"],
        is_rgb=True,
    )
    win = himena_ui.add_data_model(model)
    assert isinstance(view := win.widget, QImageView)
    assert view._is_rgb
    himena_ui.exec_action("builtins:image:split-channels")
    assert len(himena_ui.tabs.current()) == 4

def test_scale_bar(himena_ui: MainWindow):
    win = himena_ui.add_data_model(
        create_image_model(
            np.zeros((100, 100)),
            axes=[
                DimAxis(name="y", scale=0.42, unit="um"),
                DimAxis(name="x", scale=0.28, unit="um")
            ]
        )
    )
    himena_ui.exec_action("builtins:image:setup-image-scale-bar", with_params={})
    himena_ui.show()
    win.size = (300, 300)
    win.size = (200, 200)
    assert isinstance(win.widget, QImageView)
    win.widget._img_view.move_items_by(2, 2)

def test_find_nice_position():
    from himena_builtins.qt.widgets._image_components._mouse_events import _find_nice_position, _find_nice_rect_position

    for angle in np.linspace(0, np.pi * 2, 30):
        x = float(np.sin(angle))
        y = float(np.cos(angle))
        p = _find_nice_position(QtCore.QPointF(x, y), QtCore.QPointF(0, 0))
        ang_out = np.arctan2(p.y(), p.x())
        assert np.rad2deg(ang_out) % 45 < 0.1
        assert abs(angle - ang_out) <= 22.5

        p = _find_nice_rect_position(QtCore.QPointF(x, y), QtCore.QPointF(0, 0))
        assert abs(p.x()) == abs(p.y())

def test_flat_roi_always_selected(qtbot: QtBot):
    view = QImageView()
    qtbot.addWidget(view)
    cur_roi = LineRoi(start=(2, 4), end=(5, 5))
    view.update_model(
        create_image_model(
        np.zeros((5, 10, 10)),
            axes=[
                DimAxis(name="t"),
                DimAxis(name="y"),
                DimAxis(name="x"),
            ],
            current_roi=cur_roi,
            rois=RoiListModel(
                items=[cur_roi], indices=np.array([[-1, -1]]), axis_names=["t"],
            ),
        )
    )
    assert view._img_view._current_roi_item is not None
    view._dims_slider.setValue((1,))
    assert view._img_view._current_roi_item is not None

def test_select_rois(himena_ui: MainWindow):
    model = create_image_model(
        np.zeros((4, 4, 10, 10)),
        axes=["t", "z", "y", "x"],
        rois=RoiListModel(
            items=[
                LineRoi(name="ROI-0", start=(1, 1), end=(4, 5)),
                PointRoi2D(name="ROI-1", x=1, y=5),
            ],
            indices=np.array([[0, 0], [0, 0]], dtype=np.int32),
        )
    )
    himena_ui.add_data_model(model)
    himena_ui.exec_action("builtins:select-image-rois", with_params={"selections": [0]})
    assert isinstance(lmodel := himena_ui.current_model.value, RoiListModel)
    assert len(lmodel) == 1
    assert lmodel.items[0].name == "ROI-0"

def test_handle_events(qtbot: QtBot):
    view = QImageView()
    qtbot.addWidget(view)
    handles = RoiSelectionHandles(view._img_view)
    # line
    line_roi = _rois.QLineRoi(0, 0, 5, 5)
    handles.connect_roi(line_roi)
    for h in handles._handles:
        h.moved_by_mouse.emit(QtCore.QPointF(1, 1), QtCore.QPointF(0, 0))
    line_roi.setLine(0, 0, 6, 6)
    # rect
    rect_roi = _rois.QRectangleRoi(0, 0, 5, 8)
    handles.connect_roi(rect_roi)
    for h in handles._handles:
        h.moved_by_mouse.emit(QtCore.QPointF(1, 1), QtCore.QPointF(0, 0))
    rect_roi.setRect(0, 0, 6, 9)
    # path
    path_roi = _rois.QPolygonRoi([0, 1, 3, 0], [3, 5, 5, 3])
    handles.connect_roi(path_roi)
    for h in handles._handles:
        h.moved_by_mouse.emit(QtCore.QPointF(1, 1), QtCore.QPointF(0, 0))
    # circle
    circle_roi = _rois.QCircleRoi(0, 0, 5)
    handles.connect_roi(circle_roi)
    for h in handles._handles:
        h.moved_by_mouse.emit(QtCore.QPointF(1, 1), QtCore.QPointF(0, 0))
    # point
    point_roi = _rois.QPointRoi(0, 0)
    handles.connect_roi(point_roi)
    for h in handles._handles:
        h.moved_by_mouse.emit(QtCore.QPointF(1, 1), QtCore.QPointF(0, 0))
    point_roi.setPos(1, 1)
    # points
    points_roi = _rois.QPointsRoi([0, 1], [0, 1])
    handles.connect_roi(points_roi)
    for h in handles._handles:
        h.moved_by_mouse.emit(QtCore.QPointF(1, 1), QtCore.QPointF(0, 0))
    points_roi.setPos(1, 1)
    # rotated rect
    rotated_roi = _rois.QRotatedRectangleRoi(
        QtCore.QPointF(0, 0),
        QtCore.QPointF(10, 10),
        6,
    )
    handles.connect_roi(rotated_roi)
    for h in handles._handles:
        h.moved_by_mouse.emit(QtCore.QPointF(1, 1), QtCore.QPointF(0, 0))

def test_play(qtbot: QtBot):
    view = QImageView()
    qtbot.addWidget(view)
    view.update_model(
        create_image_model(
            np.zeros((4, 10, 10)),
            axes=[
                DimAxis(name="t"),
                DimAxis(name="y"),
                DimAxis(name="x"),
            ],
        )
    )
    tslider = view._dims_slider._sliders[0]
    tslider._make_context_menu_for_play_btn()
    tslider._on_play_timer_timeout()  # nothing happens because button is not checked
    # test "once"
    tslider._slider.setValue(0)
    tslider._play_back_mode = "once"
    tslider._play_btn.click()
    for _ in range(4):
        tslider._on_play_timer_timeout()
    assert tslider._slider.value() == 3
    assert not tslider._play_timer.isActive()
    # test "loop"
    tslider._slider.setValue(0)
    tslider._play_back_mode = "loop"
    tslider._play_btn.click()
    for _ in range(4):
        tslider._on_play_timer_timeout()
    assert tslider._slider.value() == 0
    assert tslider._play_timer.isActive()
    tslider._stop_play()
    # test "pingpong"
    tslider._slider.setValue(0)
    tslider._play_back_mode = "pingpong"
    tslider._play_btn.click()
    for _ in range(4):
        tslider._on_play_timer_timeout()
    assert tslider._slider.value() == 2
    assert tslider._play_increment == -1
    for _ in range(3):
        tslider._on_play_timer_timeout()
    assert tslider._play_increment == 1
    assert tslider._slider.value() == 1
    tslider._stop_play()

def test_propagate(himena_ui: MainWindow):
    win0 = himena_ui.add_data_model(
        create_image_model(
            np.zeros((3, 10, 12), dtype=np.uint8),
            axes=["c", "y", "x"],
            channel_axis=0,
            channels=["red", "green", "blue"],
        )
    )
    win1 = himena_ui.add_data_model(
        create_image_model(
            np.zeros((3, 10, 14), dtype=np.uint8),
            axes=["c", "y", "x"],
            channel_axis=0,
        )
    )
    win2 = himena_ui.add_data_model(
        create_image_model(
            np.zeros((3, 10, 12), dtype=np.uint8),
            axes=["t", "y", "x"],
        )
    )
    win3 = himena_ui.add_data_model(
        create_image_model(
            np.zeros((2, 3, 10, 12), dtype=np.uint8),
            axes=["t", "c", "y", "x"],
            channel_axis=1,
        )
    )

    himena_ui.exec_action("builtins:image:propagate-colormaps", window_context=win0)

    def get_colormap_names(win: SubWindow) -> list[str | None]:
        return [c.colormap for c in win.to_model().metadata.channels]

    assert get_colormap_names(win0) == ["cmap:red", "cmap:green", "cmap:blue"]
    assert get_colormap_names(win1) == ["cmap:red", "cmap:green", "cmap:blue"]
    assert get_colormap_names(win2) == ["matlab:gray"]
    assert get_colormap_names(win3) == ["cmap:red", "cmap:green", "cmap:blue"]
    himena_ui.current_window = win2
    himena_ui.exec_action("builtins:image:set-channel-axis", with_params={"axis": 0})

def test_image_widget_hover_info(qtbot: QtBot):
    image_view = QImageView()
    qtbot.addWidget(image_view)
    rng = np.random.default_rng(0)
    with WidgetTester(image_view) as tester:
        tester.update_model(value=rng.random((20, 20)))
        tester.widget._on_hovered(QtCore.QPointF(2.1, 2.5))
        tester.widget._on_hovered(QtCore.QPointF(30, 40))  # out of range
        tester.update_model(
            create_image_model(
                rng.random((20, 20)),
                axes=[
                    DimAxis(name="y", scale=0.1, unit="mm"),
                    DimAxis(name="x", scale=0.1, unit="mm"),
                ]
            )
        )
        tester.widget._on_hovered(QtCore.QPointF(2.1, 2.5))
        tester.widget._on_hovered(QtCore.QPointF(30, 40))  # out of range

        # FFT
        tester.update_model(
            create_image_model(
                np.fft.fftn(rng.random((20, 20))),
            ).astype(StandardType.IMAGE_FOURIER)
        )
        tester.widget._on_hovered(QtCore.QPointF(2.1, 2.5))
        tester.widget._on_hovered(QtCore.QPointF(30, 40))  # out of range
        tester.update_model(
            create_image_model(
                np.fft.fftn(rng.random((20, 20))),
                axes=[
                    DimAxis(name="y", scale=0.1, unit="mm"),
                    DimAxis(name="x", scale=0.1, unit="mm"),
                ]
            ).astype(StandardType.IMAGE_FOURIER)
        )
        tester.widget._on_hovered(QtCore.QPointF(2.1, 2.5))
        tester.widget._on_hovered(QtCore.QPointF(30, 40))  # out of range

def test_scale_bar_widget():
    from himena_builtins.qt.widgets._image_commands import ScaleBarSpecWidget

    widget = ScaleBarSpecWidget()
    widget.get_value()
    widget.set_value(
        {
            "shape": (3.0, 1.0),
            "color": "#FF0000",
            "anchor_pos": "bottom-left",
            "offset": (2, 2),
        }
    )
    widget.get_value()



from ._selection_range_edit import QSelectionRangeEdit
from ._base import QTableBase, Editability, FLAGS, parse_string
from ._formatter import format_table_value
from ._header import QHorizontalHeaderView, QVerticalHeaderView
from ._selection_model import SelectionModel

__all__ = [
    "QSelectionRangeEdit",
    "QTableBase",
    "FLAGS",
    "parse_string",
    "Editability",
    "format_table_value",
    "QHorizontalHeaderView",
    "QVerticalHeaderView",
    "SelectionModel",
]



from __future__ import annotations
from typing import Any, cast

from magicgui.widgets.bases import ValueWidget
from magicgui.types import Undefined
from magicgui.application import use_app
from magicgui.backends._qtpy.widgets import QBaseValueWidget
from cmap import Color, Colormap
from superqt import QColormapComboBox
from himena.utils.misc import lru_cache
from himena.qt._qcoloredit import QColorEdit


class _QColorEdit(QBaseValueWidget):
    _qwidget: QColorEdit

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            QColorEdit,
            "rgba",
            "setRgba",
            "colorChanged",
            **kwargs,
        )

    def _pre_set_hook(self, value: Any) -> Any:
        return Color(value)

    def _post_get_hook(self, value: Any) -> Any:
        return Color(value)


class ColorEdit(ValueWidget[Color]):
    def __init__(self, value=Undefined, **kwargs):
        app = use_app()
        assert app.native

        ValueWidget.__init__(
            self,
            value=value,
            widget_type=_QColorEdit,
            **kwargs,
        )


@lru_cache
def get_colormaps(category: str = "qualitative") -> list[str]:
    cat = Colormap.catalog()
    return sorted(cat.unique_keys(categories=[category]))


class _ColormapEdit(QBaseValueWidget):
    _qwidget: QColormapComboBox

    def __init__(self, **kwargs: Any) -> None:
        cat = kwargs.pop("category", None)
        defaults = kwargs.pop("defaults", None)

        super().__init__(
            QColormapComboBox,
            "currentColormap",
            "setCurrentColormap",
            "currentColormapChanged",
            **kwargs,
        )
        qwidget = cast(QColormapComboBox, self._qwidget)
        if defaults:
            qwidget.addColormaps(defaults)
        if cat:
            qwidget.addColormaps(get_colormaps(cat))
        qwidget.setMinimumHeight(22)
        qwidget.setUserAdditionsAllowed(True)

    def _pre_set_hook(self, value: Any) -> Any:
        return Colormap(value)

    def _post_get_hook(self, value: Any) -> Any:
        return Colormap(value)


class ColormapEdit(ValueWidget[Colormap]):
    def __init__(self, value=Undefined, **kwargs):
        app = use_app()
        assert app.native

        if category := kwargs.pop("category", None):
            backend_kwargs = {"category": category}
        elif defaults := kwargs.pop("defaults", None):
            backend_kwargs = {"defaults": defaults}
        else:
            backend_kwargs = {}
        ValueWidget.__init__(
            self,
            value=value,
            widget_type=_ColormapEdit,
            backend_kwargs=backend_kwargs,
            **kwargs,
        )



from __future__ import annotations
from typing import Any, Literal

import numpy as np
from pydantic_compat import Field
from himena.standards.plotting.components import BasePlotModel, Face, Edge


class PlotModelXY(BasePlotModel):
    x: Any = Field(..., description="X-axis values.")
    y: Any = Field(..., description="Y-axis values.")


class Scatter(PlotModelXY):
    """Plot model for scatter plot."""

    symbol: Any | None = Field(None, description="Symbol of the markers.")
    size: Any | None = Field(None, description="Size of the markers.")
    face: Face = Field(
        default_factory=Face, description="Properties of the marker faces."
    )
    edge: Edge = Field(
        default_factory=Edge, description="Properties of the marker edges."
    )

    def plot_option_dict(self) -> dict[str, Any]:
        from himena.qt.magicgui import EdgePropertyEdit, FacePropertyEdit

        return {
            "name": {"widget_type": "LineEdit", "value": self.name},
            "symbol": {"widget_type": "LineEdit", "value": self.symbol},
            "size": {"annotation": float, "value": self.size},
            "face": {"widget_type": FacePropertyEdit, "value": self.face.model_dump()},
            "edge": {"widget_type": EdgePropertyEdit, "value": self.edge.model_dump()},
        }


class Line(PlotModelXY):
    """Plot model for line plot."""

    edge: Edge = Field(default_factory=Edge, description="Properties of the line.")
    marker: Scatter | None = Field(None, description="Marker of the line.")

    def plot_option_dict(self) -> dict[str, Any]:
        from himena.qt.magicgui import EdgePropertyEdit

        return {
            "name": {"widget_type": "LineEdit", "value": self.name},
            "edge": {"widget_type": EdgePropertyEdit, "value": self.edge.model_dump()},
        }


class Bar(PlotModelXY):
    """Plot model for bar plot."""

    bottom: float | Any = Field(0, description="Bottom values of the bars.")
    bar_width: float | None = Field(None, description="Width of the bars.")
    orient: Literal["vertical", "horizontal"] = Field(
        "vertical", description="Orientation of the bar plots."
    )
    face: Face = Field(default_factory=Face, description="Properties of the bars.")
    edge: Edge = Field(default_factory=Edge, description="Properties of the bars.")

    def plot_option_dict(self) -> dict[str, Any]:
        from himena.qt.magicgui import EdgePropertyEdit, FacePropertyEdit

        return {
            "name": {"widget_type": "LineEdit", "value": self.name},
            "bar_width": {"annotation": float, "value": self.bar_width},
            "orient": {"choices": ["vertical", "horizontal"], "value": self.orient},
            "face": {"widget_type": FacePropertyEdit, "value": self.face.model_dump()},
            "edge": {"widget_type": EdgePropertyEdit, "value": self.edge.model_dump()},
        }


class ErrorBar(PlotModelXY):
    """Plot model for error bar plot."""

    x_error: Any | None = Field(None, description="X-axis error values.")
    y_error: Any | None = Field(None, description="Y-axis error values.")
    capsize: float | None = Field(None, description="Cap size of the error bars.")
    edge: Edge = Field(
        default_factory=Edge, description="Properties of the error bars."
    )

    def plot_option_dict(self) -> dict[str, Any]:
        from himena.qt.magicgui import EdgePropertyEdit

        return {
            "name": {"widget_type": "LineEdit", "value": self.name},
            "capsize": {"annotation": float, "value": self.capsize},
            "edge": {"widget_type": EdgePropertyEdit, "value": self.edge.model_dump()},
        }


class Band(BasePlotModel):
    """Plot model for band plot."""

    x: Any = Field(..., description="X-axis values.")
    y0: Any = Field(..., description="Y-axis values of the lower bound.")
    y1: Any = Field(..., description="Y-axis values of the upper bound.")
    orient: Literal["vertical", "horizontal"] = Field(
        "vertical", description="Orientation of the band fill."
    )
    face: Face = Field(default_factory=Face, description="Properties of the band fill.")
    edge: Edge = Field(default_factory=Edge, description="Properties of the band edge.")

    def plot_option_dict(self) -> dict[str, Any]:
        from himena.qt.magicgui import EdgePropertyEdit, FacePropertyEdit

        return {
            "name": {"widget_type": "LineEdit", "value": self.name},
            "orient": {"choices": ["vertical", "horizontal"], "value": self.orient},
            "face": {"widget_type": FacePropertyEdit, "value": self.face.model_dump()},
            "edge": {"widget_type": EdgePropertyEdit, "value": self.edge.model_dump()},
        }


class Span(BasePlotModel):
    """Plot model for span plot."""

    start: float = Field(..., description="Starting value of the lower bound.")
    end: float = Field(..., description="Ending value of the upper bound.")
    orient: Literal["vertical", "horizontal"] = Field(
        "vertical",
        description="Orientation of the span. 'vertical' means the span"
        "is vertically unlimited.",
    )
    face: Face = Field(default_factory=Face, description="Properties of the span fill.")
    edge: Edge = Field(default_factory=Edge, description="Properties of the span edge.")

    def plot_option_dict(self) -> dict[str, Any]:
        from himena.qt.magicgui import EdgePropertyEdit, FacePropertyEdit

        return {
            "name": {"widget_type": "LineEdit", "value": self.name},
            "x0": {"annotation": float, "value": self.start},
            "x1": {"annotation": float, "value": self.end},
            "orient": {"choices": ["vertical", "horizontal"], "value": self.orient},
            "face": {"widget_type": FacePropertyEdit, "value": self.face.model_dump()},
            "edge": {"widget_type": EdgePropertyEdit, "value": self.edge.model_dump()},
        }


class Histogram(BasePlotModel):
    """Plot model for a histogram."""

    height: Any = Field(..., description="Count or frequency values.")
    bins: Any = Field(..., description="Bin edges.")
    orient: Literal["vertical", "horizontal"] = Field(
        "vertical", description="Orientation of the histogram."
    )
    face: Face = Field(
        default_factory=Face, description="Properties of the histogram face."
    )
    edge: Edge = Field(
        default_factory=Edge, description="Properties of the histogram edge."
    )

    def plot_option_dict(self) -> dict[str, Any]:
        from himena.qt.magicgui import EdgePropertyEdit, FacePropertyEdit

        return {
            "name": {"widget_type": "LineEdit", "value": self.name},
            "orient": {"choices": ["vertical", "horizontal"], "value": self.orient},
            "face": {"widget_type": FacePropertyEdit, "value": self.face.model_dump()},
            "edge": {"widget_type": EdgePropertyEdit, "value": self.edge.model_dump()},
        }

    def to_band(self) -> Band:
        """Convert the histogram to a band plot."""
        x = np.repeat(self.bins, 3)[1:-1]
        y = np.zeros_like(x)
        y[1::3] = self.height
        y[2::3] = self.height
        return Band(
            x=x,
            y0=y,
            y1=np.zeros_like(y),
            orient=self.orient,
            face=self.face,
            edge=self.edge,
        )


ANCHOR_STRINGS = Literal[
    "center", "left", "right", "top", "bottom", "top_left", "top_right",
    "bottom_left", "bottom_right"
]  # fmt: skip


class Texts(PlotModelXY):
    """Plot model for texts."""

    texts: Any = Field(..., description="Texts to be displayed.")
    size: int = Field(12, description="Font size of the texts.")
    color: str = Field("black", description="Font color of the texts.")
    family: str = Field("Arial", description="Font family of the texts.")
    anchor: ANCHOR_STRINGS = Field(
        "center",
        description="Anchor position of the texts. 'center' means the center of the text.",
    )
    rotation: float = Field(0, description="Rotation angle of the texts.")

    def plot_option_dict(self) -> dict[str, Any]:
        return {
            "name": {"widget_type": "LineEdit", "value": self.name},
            "font_size": {"annotation": int, "value": self.size},
            "font_color": {"annotation": str, "value": self.color},
            "font_family": {"annotation": str, "value": self.family},
        }



from himena import StandardType
from himena.plugins import register_widget_class
from himena_builtins.qt.widgets.dict_subtypes import QDataFrameStack, QArrayStack
from himena_builtins.qt.widgets.excel import QExcelEdit

register_widget_class(StandardType.DATAFRAMES, QDataFrameStack, priority=50)
register_widget_class(StandardType.ARRAYS, QArrayStack, priority=50)
register_widget_class(StandardType.EXCEL, QExcelEdit, priority=50)



from himena.qt.magicgui._plot_elements import AxisPropertyEdit

def test_axis_property_edit():
    """Test the AxisPropertyEdit widget."""
    widget = AxisPropertyEdit()
    widget.value
    widget.value = {
        "lim": (2, 5),
        "scale": "linear",
        "label": "X-axis",
        "grid": True,
    }



from __future__ import annotations

import numpy as np

from himena_builtins.qt.widgets.array import QArrayView, QArrayViewControl
from himena_builtins.qt.widgets.dataframe import QDataFrameView, QDataFrameViewControl
from himena_builtins.qt.widgets.dict import QDictOfWidgetEdit, QTabControl
from himena.types import WidgetDataModel
from himena.consts import StandardType
from himena.plugins import validate_protocol


class QDataFrameStack(QDictOfWidgetEdit):
    """DataFrame Stack widget."""

    __himena_widget_id__ = "builtins:QDataFrameStack"
    __himena_display_name__ = "Built-in DataFrame Stack"

    def __init__(self):
        super().__init__()
        self._model_type_component = StandardType.DATAFRAME
        self._model_type = StandardType.DATAFRAMES
        self._control = QDataFramesControl()

    def _default_widget(self) -> QDataFrameView:
        table = QDataFrameView()
        table.update_model(WidgetDataModel(value={}, type=StandardType.DATAFRAME))
        return table

    @validate_protocol
    def control_widget(self) -> QDataFramesControl:
        return self._control


class QDataFramesControl(QDataFrameViewControl, QTabControl):
    def update_for_component(self, widget):
        return self.update_for_table(widget)


class QArrayStack(QDictOfWidgetEdit):
    __himena_widget_id__ = "builtins:QArrayStack"
    __himena_display_name__ = "Built-in Array Stack"

    def __init__(self):
        super().__init__()
        self._model_type_component = StandardType.ARRAY
        self._model_type = StandardType.ARRAYS
        self._control = QArrayStackControl()

    def _default_widget(self) -> QArrayView:
        view = QArrayView()
        view.update_model(
            WidgetDataModel(value=np.zeros((0, 0)), type=StandardType.ARRAY)
        )
        return view

    @validate_protocol
    def control_widget(self) -> QArrayStackControl:
        return self._control


class QArrayStackControl(QArrayViewControl, QTabControl):
    def update_for_component(self, widget: QArrayView):
        return self.update_for_array(widget)
